#: The column name in a dataframe parsed with :meth:`~.OntologyParser.parse_to_dataframe`
#: for the column of the entity's default/preferred label
DEFAULT_LABEL = "default_label"
#: The column name for the id of each entity
IDX = "idx"
#: The column name for the synonyms/alternative labels for each entity
SYN = "syn"
#: The column name for the type of mapping from default label to synonym - e.g. xref, exactSyn etc. Usually defined by the ontology
MAPPING_TYPE = "mapping_type"
#: The origin of a dataset - e.g. HGNC release 2.1, MEDDRA 24.1 etc.
#: Note, this is different from the parser.name, as is used to identify the origin of a mapping back to a data source
DATA_ORIGIN = "data_origin"



import dataclasses

import pandas as pd
import streamlit as st
from hydra.utils import instantiate
from kazu.data import MentionConfidence, OntologyStringResource
from kazu.krt.resource_manager import ResourceManager
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.krt import load_config


@st.cache_resource(show_spinner="Loading parsers")
def load_parsers() -> list[OntologyParser]:
    cfg = load_config()
    parsers = []
    for parser in cfg.ontologies.parsers.values():
        parsers.append(instantiate(parser))
    return parsers


@st.cache_resource(show_spinner="Building the ResourceManager")
def get_resource_manager() -> ResourceManager:
    return ResourceManager(load_parsers())


def create_new_resource_with_updated_synonyms(
    new_conf: MentionConfidence, new_cs: bool, resource: OntologyStringResource
) -> OntologyStringResource:
    """Create a new :class:`.OntologyStringResource` with updated :class:`.Synonym`\\s.

    This function takes a :class:`.MentionConfidence`\\, a boolean value, and an
    :class:`.OntologyStringResource` object as inputs. It creates a new :class:`.OntologyStringResource`
    object with updated synonyms based on the provided :class:`.MentionConfidence` and boolean
    values. The new_conf parameter represents the new :class:`.MentionConfidence` to be set for
    the :class:`.Synonym`\\s. The new_cs parameter represents whether the :class:`.Synonym`\\s should be case
    sensitive or not. The resource parameter is the original :class:`.OntologyStringResource`
    object whose :class:`.Synonym`\\s are to be updated.

    :param new_conf:
    :param new_cs:
    :param resource:
    :return:
    """
    new_original_syns = set()
    for syn in resource.original_synonyms:
        new_original_syns.add(
            dataclasses.replace(syn, case_sensitive=new_cs, mention_confidence=new_conf)
        )
    new_alternative_syns = set()
    for syn in resource.alternative_synonyms:
        new_alternative_syns.add(
            dataclasses.replace(syn, case_sensitive=new_cs, mention_confidence=new_conf)
        )

    return dataclasses.replace(
        resource,
        original_synonyms=frozenset(new_original_syns),
        alternative_synonyms=frozenset(new_alternative_syns),
    )


def resource_to_df(
    resource: OntologyStringResource, include_behaviour: bool = False
) -> pd.DataFrame:
    """Convert an :class:`.OntologyStringResource` to a :class:`~pandas.DataFrame` for
    display in Streamlit.

    This function takes an OntologyStringResource object as input and transforms it into a pandas DataFrame.
    The DataFrame has four columns: `type`, `text`, `confidence`, and `case_sensitive`. The `type` column
    indicates whether the synonym is "original" or "alternative". The `text` column contains the synonym text.
    The `confidence` column contains the mention confidence, and the `case_sensitive` column indicates
    whether the synonym is case sensitive or not.

    :param resource:
    :param include_behaviour: if True, include the behaviour column in the DataFrame
    :return:
    """
    data = [
        (
            "original",
            syn.text,
            syn.mention_confidence.name,
            syn.case_sensitive,
        )
        for syn in resource.original_synonyms
    ] + [
        (
            "alternative",
            syn.text,
            syn.mention_confidence.name,
            syn.case_sensitive,
        )
        for syn in resource.alternative_synonyms
    ]
    df = pd.DataFrame.from_records(data, columns=["type", "text", "confidence", "case_sensitive"])
    if include_behaviour:
        df["behaviour"] = resource.behaviour
    return df



import logging
from collections import defaultdict
from collections.abc import Iterable

import ahocorasick
from kazu.data import Document, Entity, MentionConfidence, LinkingMetrics, LinkingCandidate
from kazu.database.in_memory_db import SynonymDatabase, ParserName, NormalisedSynonymStr
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.steps import document_iterating_step, ParserDependentStep
from kazu.utils.caching import kazu_disk_cache
from kazu.utils.grouping import sort_then_group
from kazu.utils.spacy_pipeline import SpacyPipelines, basic_spacy_pipeline, BASIC_PIPELINE_NAME

logger = logging.getLogger(__name__)

EntityClass = str
CaseSensitive = bool
EntityInfoToOntologyInfoMapping = defaultdict[
    tuple[EntityClass, MentionConfidence, CaseSensitive, NormalisedSynonymStr, str], set[ParserName]
]


class MemoryEfficientStringMatchingStep(ParserDependentStep):
    """A wrapper for the ahocorasick algorithm.

    In testing, this implementation is comparable in speed to a spaCy
    `PhraseMatcher <https://spacy.io/api/phrasematcher>`_\\,
    and uses a fraction of the memory. Since this implementation is unaware
    of NLP concepts such as tokenization, we backfill this capability by
    checking for word boundaries with a custom spaCy tokenizer.
    """

    def __init__(self, parsers: Iterable[OntologyParser]):
        super().__init__(parsers)
        self.parsers = parsers
        self.automaton = self._create_automaton()
        self.spacy_pipelines = SpacyPipelines()
        self.spacy_pipelines.add_from_func(BASIC_PIPELINE_NAME, basic_spacy_pipeline)
        self.synonym_db = SynonymDatabase()

    @kazu_disk_cache.memoize(ignore={0})
    def _create_automaton(self) -> ahocorasick.Automaton:
        """Create `ahocorasick.Automaton
        <https://pyahocorasick.readthedocs.io/en/latest/#automaton-class>`_\\ s
        for parsers."""
        key_to_ontology_info: defaultdict[str, EntityInfoToOntologyInfoMapping] = defaultdict(
            lambda: defaultdict(set)
        )

        logger.info("Creating ahocorasick Automaton")
        for parser in self.parsers:
            parser_resources = parser.populate_databases(return_resources=True)
            if parser_resources is None:
                logger.warning(
                    "tried to create ahocorasick data for parser %s, but no ontology resources are available",
                    parser.name,
                )
                continue
            if len(parser_resources) == 0:
                logger.warning(
                    "tried to create ahocorasick data for parser %s, but no ontology resources were produced",
                    parser.name,
                )
                continue

            for resource in parser_resources:
                # a resource can have different syn_norms for different parsers,
                # since the string normalizer's output depends on the entity class.
                # Also, a synonym may exist in multiple LinkingCandidate.raw_synonyms
                syn_norm = resource.syn_norm_for_linking(parser.entity_class)
                for syn in resource.active_ner_synonyms():
                    entity_key = (
                        parser.entity_class,
                        syn.mention_confidence,
                        syn.case_sensitive,
                        syn_norm,
                        syn.text,
                    )
                    key_to_ontology_info[syn.text.lower()][entity_key].add(parser.name)

        if len(key_to_ontology_info) > 0:
            case_insensitive_automaton = ahocorasick.Automaton()
            for key, entity_to_ontology_info in key_to_ontology_info.items():
                case_insensitive_automaton.add_word(key, entity_to_ontology_info)
            case_insensitive_automaton.make_automaton()
        else:
            raise RuntimeError(
                f"No valid resources were available for {self.__class__.__name__}. Either remove this step "
                "from your pipeline or ensure at least one valid resource is generated from the chosen "
                "parsers."
            )

        return case_insensitive_automaton

    def _word_is_valid(
        self, start_char: int, end_char: int, starts: set[int], ends: set[int]
    ) -> bool:
        return start_char in starts and end_char in ends

    def _case_matches(self, actual_match: str, original_casing: str, case_sensitive: bool) -> bool:
        # no need to check the strings actually match if it's not case sensitive
        # as this is already done as part of the string matching, so we know that
        # actual_match.lower() == original_casing.lower()
        return (not case_sensitive) or actual_match == original_casing

    def _process_automaton(
        self,
        automaton: ahocorasick.Automaton,
        matchable_text: str,
        original_text: str,
        starts: set[int],
        ends: set[int],
    ) -> list[Entity]:
        entities = []
        for end_index, ontology_dict in automaton.iter(matchable_text):
            # all keys should have the same value for synonym text
            match_key = next(iter(ontology_dict.keys()))[-1]
            start_index = end_index - len(match_key) + 1
            matched_text = original_text[start_index : end_index + 1]
            if self._word_is_valid(start_index, end_index, starts, ends):

                for entity_class, entity_info_groups in sort_then_group(
                    ontology_dict.keys(), key_func=lambda x: x[0]
                ):
                    candidates: set[LinkingCandidate] = set()
                    confidences: defaultdict[str, set[MentionConfidence]] = defaultdict(set)
                    for entity_info in entity_info_groups:
                        (
                            _,
                            confidence,
                            case_sensitive,
                            syn_norm,
                            original_case,
                        ) = entity_info

                        # filter out cases where we've specified we only want exact matches, but we only have a lowercase match
                        if not self._case_matches(
                            actual_match=matched_text,
                            original_casing=original_case,
                            case_sensitive=case_sensitive,
                        ):
                            continue

                        parser_name_set = ontology_dict[entity_info]

                        for parser_name in parser_name_set:
                            confidences[parser_name].add(confidence)
                            candidates.add(self.synonym_db.get(parser_name, syn_norm))

                    if len(candidates) > 0:

                        confidence_values = {max(conf_set) for conf_set in confidences.values()}
                        chosen_conf = max(confidence_values)
                        if len(confidence_values) > 1:
                            logger.warning(
                                f"confidences conflict between parsers for {matched_text}: {confidences}. The maximum will be selected ({chosen_conf})"
                            )

                        e = Entity.load_contiguous_entity(
                            start=start_index,
                            end=end_index + 1,
                            match=matched_text,
                            entity_class=entity_class,
                            namespace=self.namespace(),
                            mention_confidence=chosen_conf,
                        )
                        for candidate in candidates:
                            e.add_or_update_linking_candidate(
                                candidate, LinkingMetrics(exact_match=True)
                            )
                        entities.append(e)
        return entities

    @document_iterating_step
    def __call__(self, doc: Document) -> None:

        for section in doc.sections:
            spacy_doc = self.spacy_pipelines.process_single(section.text, BASIC_PIPELINE_NAME)
            starts, ends = set(), set()
            for tok in spacy_doc:
                starts.add(tok.idx)
                ends.add(tok.idx + len(tok) - 1)

            section.entities.extend(
                self._process_automaton(
                    self.automaton, section.text.lower(), section.text, starts, ends
                )
            )



import streamlit as st
from kazu.data import MentionConfidence
from kazu.krt.components import (
    ResourceEditor,
)
from kazu.krt.resource_discrepancy_editor.utils import (
    SynonymDiscrepancy,
    ResourceDiscrepancyManger,
)
from kazu.krt.utils import create_new_resource_with_updated_synonyms, get_resource_manager


@st.cache_resource
def get_resource_merge_manager(parser_name: str) -> ResourceDiscrepancyManger:
    manager = get_resource_manager()
    return ResourceDiscrepancyManger(
        parser_name=parser_name,
        manager=manager,
    )


def reset() -> None:
    # type ignore as mypy doesn't pick up annotations
    get_resource_manager.clear()  # type: ignore[attr-defined]
    get_resource_merge_manager.clear()  # type: ignore[attr-defined]


def get_resource_merge_manager_for_parser(parser_name: str) -> ResourceDiscrepancyManger:
    return get_resource_merge_manager(parser_name=parser_name)


class ResourceDiscrepancyResolutionForm:
    """This class is used to handle the resolution of resource discrepancies.

    It provides a form interface for the user to interact with and resolve
    discrepancies.
    """

    ATTEMPT_AUTOFIX = "ATTEMPT_AUTOFIX"

    def __init__(self, manager: ResourceDiscrepancyManger):
        self.manager = manager
        self.autofix_already_attempted = False

    @staticmethod
    def state_key(parser_name: str) -> str:
        return f"{ResourceDiscrepancyResolutionForm.__name__}{parser_name}"

    def display_main_form(self) -> None:
        """Display the main form for resolving resource discrepancies.

        :return:
        """
        st.write("select a row to resolve a discrepancy")
        event = st.dataframe(
            self.manager.summary_df(),
            use_container_width=True,
            selection_mode="single-row",
            on_select="rerun",
            hide_index=True,
            column_config={"id": None},
        )
        if not self.autofix_already_attempted:
            self._display_attempt_autofix_button()
            self._run_autofix_if_requested()

        for row_id in event.get("selection", {}).get("rows", []):
            self._display_discrepancy_form_for_selected_index(index=row_id)

    def _reset_form(self) -> None:
        """Reset the form by setting the AUTOFIX_ALREADY_ATTEMPTED session state to
        False."""
        self.autofix_already_attempted = False

    def _display_attempt_autofix_button(self) -> None:
        """Display a button for the user to attempt to autofix discrepancies.

        The button is disabled if autofix has already been attempted.
        """
        st.button(
            "Attempt to Autofix discrepancies",
            key=ResourceDiscrepancyResolutionForm.get_attempt_autofix_key(self.manager.parser_name),
            disabled=self.autofix_already_attempted,
        )

    @staticmethod
    def get_attempt_autofix_key(parser_name: str) -> str:
        return f"{ResourceDiscrepancyResolutionForm.ATTEMPT_AUTOFIX}_{parser_name}"

    def _run_autofix_if_requested(self) -> None:
        """If the user has requested to run autofix, apply autofix to all resources and
        rerun the script.

        :return:
        """
        if st.session_state.get(
            ResourceDiscrepancyResolutionForm.get_attempt_autofix_key(self.manager.parser_name)
        ):
            self.manager.apply_autofix_to_all()
            self.autofix_already_attempted = True
            st.rerun()

    def _submit_form_batch(
        self, conf: MentionConfidence, cs: bool, conflict: SynonymDiscrepancy, index: int
    ) -> None:
        """Submit the form to resolve a discrepancy.

        All synonyms will be updated with the provided case sensitivity and confidence.

        :param conf:
        :param cs:
        :param conflict:
        :param index:
        :return:
        """
        new_resource = create_new_resource_with_updated_synonyms(
            new_conf=conf, new_cs=cs, resource=conflict.auto_resource
        )
        self.manager.commit(
            original_human_resource=conflict.human_resource,
            new_resource=new_resource,
            index=index,
        )
        self._reset_form()

    def _submit_form_individual(self, discrepancy: SynonymDiscrepancy, index: int) -> None:
        """Submit the form to resolve a discrepancy with individual edits.

        :param discrepancy:
        :param index:
        :return:
        """

        for _, new_resource in ResourceEditor.extract_form_data_from_state(
            parser_name=self.manager.parser_name, resources={discrepancy.auto_resource}
        ):

            self.manager.commit(
                original_human_resource=discrepancy.human_resource,
                new_resource=new_resource,
                index=index,
            )
            self._reset_form()

    def _display_discrepancy_form_for_selected_index(self, index: int) -> None:
        """Display the discrepancy form for the selected index.

        :param index:
        :return:
        """
        discrepancy = self.manager.unresolved_discrepancies[index]
        st.write(discrepancy.dataframe())
        form = st.radio("select a form", options=["apply to all", "edit individual"])
        if form == "apply to all":
            self._display_batch_edit_form(discrepancy, index)
        else:
            ResourceEditor.display_resource_editor(
                resources={discrepancy.auto_resource},
                on_click_override=ResourceDiscrepancyResolutionForm._submit_form_individual,
                args=(
                    self,
                    discrepancy,
                    index,
                ),
            )

    def _display_batch_edit_form(self, discrepancy: SynonymDiscrepancy, index: int) -> None:
        """Display the batch edit form for the given discrepancy.

        :param discrepancy:
        :param index:
        :return:
        """
        with st.form("conflict_editor"):
            header = st.columns([2, 2])
            header[0].subheader("case sensitivity")
            header[1].subheader("confidence")
            row1 = st.columns([2, 2])
            defaults = next(iter(discrepancy.auto_resource.original_synonyms))
            cs, conf = ResourceEditor.display_case_sensitivity_and_confidence_selector(
                row=row1, default_syn=defaults
            )
            st.form_submit_button(
                "Submit",
                on_click=ResourceDiscrepancyResolutionForm._submit_form_batch,
                args=(
                    self,
                    conf,
                    cs,
                    discrepancy,
                    index,
                ),
            )



"""This script does the following things:

1. Query Ensembl to get gene to protein ID maps
2. Query wikidata sparql to get a list of wikidata ids to Ensembl gene or Ensembl protein IDs
3. Query wikidata api to get a list of wikipedia page urls with the wikidata IDs from 2
4. Query Wikipedia API to get page content for each page from 3
5. Join wiki page content to Ensembl gene ids based on above relationships
"""

import argparse
import dataclasses
import json
import time
from collections import defaultdict
from collections.abc import Iterable
from io import StringIO
from pathlib import Path
from typing import Any, Optional
from urllib.parse import unquote

from diskcache import Cache
from kazu.utils.caching import CacheProtocol
from pandas import Index
from tqdm import tqdm

try:
    import mwparserfromhell
except ImportError:
    raise ImportError("this script requires mwparserfromhell to be installed")

import pandas as pd
import requests

from requests.adapters import HTTPAdapter, Retry

WIKIDATA_SPARQL_URL = "https://query.wikidata.org/sparql"
WIKIDATA_API_URL = "https://www.wikidata.org/w/api.php"
WIKIPEDIA_API_URL = "https://en.wikipedia.org/w/api.php"
BIOMART_URL = "http://www.ensembl.org/biomart/martservice"

WIKIDATA_SPARQL_ENSEMBL_GENE = r"""
SELECT DISTINCT ?item ?itemLabel ?Ensembl_gene_ID WHERE {
  SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE]". }
  {
    SELECT DISTINCT ?item WHERE {
      ?item p:P594 ?statement0.
      ?statement0 ps:P594 _:anyValueP594.
    }
  }
  OPTIONAL { ?item wdt:P594 ?Ensembl_gene_ID. }
  FILTER STRSTARTS(?Ensembl_gene_ID, "ENSG")
}"""
WIKIDATA_SPARQL_ENSEMBL_PROTEIN = r"""
SELECT DISTINCT ?item ?itemLabel ?Ensembl_protein_ID WHERE {
    SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE]". }
  {
    SELECT DISTINCT ?item WHERE {
      ?item p:P705 ?statement0.
      ?statement0 ps:P705 _:anyValueP705.
    }
  }
  OPTIONAL { ?item wdt:P705 ?Ensembl_protein_ID. }
  FILTER STRSTARTS(?Ensembl_protein_ID, "ENSP")
}"""


BIOMART_GENE_TO_PROTEIN_QUERY = r"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE Query>
<Query  virtualSchemaName = "default" formatter = "CSV" header = "0" uniqueRows = "0" count = "" datasetConfigVersion = "0.6" >
    <Dataset name = "hsapiens_gene_ensembl" interface = "default" >
        <Attribute name = "ensembl_gene_id" />
        <Attribute name = "ensembl_peptide_id" />
    </Dataset>
</Query>"""

# wikidata allows up to 50 ids per request
WIKIPEDIA_API_CHUNK_SIZE = 50

WIKIPEDIA_URL_PREFIX = "https://en.wikipedia.org/wiki/"


def get_retry() -> Retry:
    return Retry(
        total=100, backoff_factor=0.2, connect=100, redirect=100, read=100, status=100, other=100
    )


cache: CacheProtocol = Cache(directory="gilda_contexts_cache")
wiki_etiquette_headers = {
    "Accept-Encoding": "gzip",
    "user_agent": "Kazu (https://github.com/AstraZeneca/KAZU)",
}


@dataclasses.dataclass
class WikipediaEnsemblMapping:
    ensembl_gene_id: str
    ensembl_protein_ids: set[str] = dataclasses.field(default_factory=set)
    wiki_gene_ids: set[str] = dataclasses.field(default_factory=set)
    wiki_protein_ids: set[str] = dataclasses.field(default_factory=set)
    wiki_gene_urls_to_text: dict[str, Optional[str]] = dataclasses.field(default_factory=dict)
    wiki_protein_urls_to_text: dict[str, Optional[str]] = dataclasses.field(default_factory=dict)

    def __hash__(self) -> int:
        return hash(id(self))

    def get_context(self) -> Optional[str]:
        items = []
        for x in self.wiki_gene_urls_to_text.values():
            if x is not None:
                items.append(x)
        for x in self.wiki_protein_urls_to_text.values():
            if x is not None:
                items.append(x)
        if len(items) > 0:
            return "\n****\n".join(items)
        return None


@cache.memoize(ignore={1})
def get_sparql_df(query: str, proxies: dict[str, str]) -> pd.DataFrame:
    print(f"downloading wikidata ids for \n{query}")
    r = requests.get(
        WIKIDATA_SPARQL_URL,
        params={"query": query},
        headers={"Accept": "text/csv"},
        proxies=proxies,
    )
    return pd.read_csv(StringIO(r.content.decode("utf-8")))


@cache.memoize(ignore={0})
def get_biomart_gene_to_protein(proxies: dict[str, str]) -> pd.DataFrame:
    print("downloading Ensembl gene protein mappings")
    r = requests.get(BIOMART_URL, params={"query": BIOMART_GENE_TO_PROTEIN_QUERY}, proxies=proxies)
    df = pd.read_csv(StringIO(r.text))
    df.columns = Index(["gene_id", "protein_id"])
    return df


def divide_chunks(items: list[str]) -> Iterable[list[str]]:
    for i in range(0, len(items), WIKIPEDIA_API_CHUNK_SIZE):
        yield items[i : i + WIKIPEDIA_API_CHUNK_SIZE]


@cache.memoize(ignore={2})
def retry_wiki_with_maxlag(
    url: str, params: dict[str, str], proxies: dict[str, str]
) -> dict[str, Any]:
    s = requests.Session()
    s.mount("http://", HTTPAdapter(max_retries=get_retry()))
    count = 1
    backoff = 2
    while True:
        response = s.get(url, params=params, proxies=proxies, headers=wiki_etiquette_headers)
        json_response: dict[str, Any] = response.json()
        error = json_response.get("error")
        if error is None:
            return json_response
        elif not error.get("code") == "maxlag":
            raise Exception(f"unknown API error: {json_response}")
        else:
            wait_for = int(response.headers.get("Retry-After", count * backoff))
            print(f"wiki api awaiting maxlag: backoff {wait_for}")
            time.sleep(wait_for)
            count += 1


@cache.memoize(ignore={0, 1, 2})
def get_wikipedia_url_from_wikidata_id(
    df_genes: pd.DataFrame, df_proteins: pd.DataFrame, proxies: dict[str, str]
) -> defaultdict[str, set[str]]:
    print("downloading wikipedia urls")
    wikidata_ids = set(df_genes["itemLabel"].tolist())
    wikidata_ids.update(df_proteins["itemLabel"].tolist())

    len_ids = len(wikidata_ids)
    len_chunks = len_ids // WIKIPEDIA_API_CHUNK_SIZE + bool(len_ids % WIKIPEDIA_API_CHUNK_SIZE)
    result = defaultdict(set)
    for chunk in tqdm(divide_chunks(sorted(wikidata_ids)), total=len_chunks):
        params = {
            "action": "wbgetentities",
            "props": "sitelinks/urls",
            "ids": "|".join(chunk),
            "format": "json",
            "maxlag": "3",
        }
        json_response = retry_wiki_with_maxlag(url=WIKIDATA_API_URL, params=params, proxies=proxies)
        entities = json_response.get("entities")
        if entities:
            for idx in wikidata_ids:
                entity = entities.get(idx)
                if entity:
                    sitelinks = entity.get("sitelinks")
                    if sitelinks:
                        # filter only the specified language
                        sitelink = sitelinks.get("enwiki")
                        if sitelink:
                            wiki_url = sitelink.get("url")
                            if wiki_url:
                                result[idx].add(unquote(wiki_url))
    return result


@cache.memoize(ignore={0, 1})
def get_wikipedia_contents_from_urls(urls: set[str], proxies: dict[str, str]) -> dict[str, str]:
    print("downloading wikipedia page contents")
    len_urls = len(urls)
    len_chunks = len_urls // WIKIPEDIA_API_CHUNK_SIZE + bool(len_urls % WIKIPEDIA_API_CHUNK_SIZE)
    # ignore these sections as they add noise
    exclude_sections = {"further reading", "references", "external links"}
    results = {}
    for chunk in tqdm(divide_chunks(sorted(urls)), total=len_chunks):

        params = {
            "action": "query",
            "prop": "revisions",
            "titles": "|".join(x.removeprefix(WIKIPEDIA_URL_PREFIX) for x in chunk),
            "rvprop": ["content"],
            "rvslots": ["main"],
            "format": "json",
            "maxlag": "3",
        }
        json_response = retry_wiki_with_maxlag(
            url=WIKIPEDIA_API_URL, params=params, proxies=proxies
        )
        query = json_response["query"]
        normalised = {x["to"]: x["from"] for x in query.get("normalized", [])}
        for page in query["pages"].values():
            title = page["title"]
            norm_title = normalised.get(title, title)
            for revision in page["revisions"]:
                main_slot = revision.get("slots", {}).get("main", {})
                wikitext = main_slot.get("*")
                if wikitext is not None:
                    parsed = mwparserfromhell.parse(wikitext)
                    sections = parsed.get_sections()
                    result = []
                    for wikisection in sections:
                        text = wikisection.strip_code()
                        text_lower = text.lower()
                        if not any(x in text_lower[:70] for x in exclude_sections):
                            result.append(text)
                    final_text = "\n***\n".join(result)
                    results[norm_title] = final_text
    return results


def create_wiki_mappings(
    gene_df: pd.DataFrame,
    protein_df: pd.DataFrame,
    ensembl_gene_to_protein_mappings: pd.DataFrame,
    wikidata_id_to_wikipedia_urls: defaultdict[str, set[str]],
    wikipage_to_text: dict[str, str],
) -> set[WikipediaEnsemblMapping]:
    mappings_by_gene: dict[str, WikipediaEnsemblMapping] = dict()
    mappings_by_protein: dict[str, WikipediaEnsemblMapping] = dict()
    print("mapping ensembl genes to proteins")
    for i, row in tqdm(
        ensembl_gene_to_protein_mappings.iterrows(), total=ensembl_gene_to_protein_mappings.shape[0]
    ):
        ensembl_gene_id = row["gene_id"]
        mapping = mappings_by_gene.get(ensembl_gene_id)
        if mapping is None:
            mapping = WikipediaEnsemblMapping(ensembl_gene_id=ensembl_gene_id)
            mappings_by_gene[ensembl_gene_id] = mapping

        ensembl_protein_id = row["protein_id"]
        # check result is not NAN or another value we don't want
        if isinstance(ensembl_protein_id, str) and ensembl_protein_id.startswith("ENSP"):
            mapping.ensembl_protein_ids.add(ensembl_protein_id)
            mappings_by_protein[ensembl_protein_id] = mapping
    print("mapping wikipedia gene text to ensembl genes")
    for i, row in tqdm(gene_df.iterrows(), total=gene_df.shape[0]):
        ensembl_id = row["Ensembl_gene_ID"]
        wiki_id = row["item"].split("/")[-1]
        mapping = mappings_by_gene.get(ensembl_id)
        if mapping is not None:
            mapping.wiki_gene_ids.add(wiki_id)
            wikipedia_urls = wikidata_id_to_wikipedia_urls.get(wiki_id, set())
            for wikipedia_url in wikipedia_urls:
                page_name = wikipedia_url.removeprefix(WIKIPEDIA_URL_PREFIX)
                mapping.wiki_gene_urls_to_text[wikipedia_url] = wikipage_to_text.get(page_name)
    print("mapping wikipedia protein text to ensembl genes")
    for i, row in tqdm(protein_df.iterrows(), total=protein_df.shape[0]):
        ensembl_id = row["Ensembl_protein_ID"]
        wiki_id = row["item"].split("/")[-1]
        mapping = mappings_by_protein.get(ensembl_id)
        if mapping is not None:
            mapping.wiki_protein_ids.add(wiki_id)
            wikipedia_urls = wikidata_id_to_wikipedia_urls.get(wiki_id, set())
            for wikipedia_url in wikipedia_urls:
                page_name = wikipedia_url.removeprefix(WIKIPEDIA_URL_PREFIX)
                mapping.wiki_protein_urls_to_text[wikipedia_url] = wikipage_to_text.get(page_name)

    return set(mappings_by_gene.values())


def extract_open_targets(path: Path, proxies: dict[str, str]) -> None:
    ensembl_gene_to_protein_mappings = get_biomart_gene_to_protein(proxies)

    gene_df_humans = get_sparql_df(WIKIDATA_SPARQL_ENSEMBL_GENE, proxies)

    protein_df_humans = get_sparql_df(WIKIDATA_SPARQL_ENSEMBL_PROTEIN, proxies)

    wikidata_id_to_wikipedia_urls = get_wikipedia_url_from_wikidata_id(
        gene_df_humans, protein_df_humans, proxies
    )

    wiki_urls = set().union(wikidata_id_to_wikipedia_urls.values())

    page_name_to_text = get_wikipedia_contents_from_urls(wiki_urls, proxies)

    mappings = create_wiki_mappings(
        gene_df=gene_df_humans,
        protein_df=protein_df_humans,
        ensembl_gene_to_protein_mappings=ensembl_gene_to_protein_mappings,
        wikidata_id_to_wikipedia_urls=wikidata_id_to_wikipedia_urls,
        wikipage_to_text=page_name_to_text,
    )

    result_dict = {}
    for mapping in mappings:
        context = mapping.get_context()
        if context is not None:
            result_dict[mapping.ensembl_gene_id] = context

    if path.exists():
        with path.open(mode="r") as f:
            existing_json_dict = json.load(f)
    else:
        existing_json_dict = {}
    existing_json_dict["OPENTARGETS_TARGET"] = result_dict

    with path.open(mode="w") as f:
        json.dump(existing_json_dict, f, sort_keys=True, indent=2)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        prog="Download contexts for gilda style disambiguation",
        description="Gilda TFIDF disambiguation requires contexts for every KB identifier, in order to disambiguate. This"
        " program queries external APIs to retrieve these",
    )
    parser.add_argument(
        "--parser_name",
        required=True,
        choices=["OPENTARGETS_TARGET"],
        help="parser to download contexts for",
    )
    parser.add_argument(
        "--output_path",
        required=True,
        type=Path,
        help="Path to write results to. If this path exists, the script will attempt to override only the contexts for the specified parser name.",
    )
    parser.add_argument(
        "--clear_cache",
        action="store_true",
        help="clear a local cache to force re-download of all data",
    )
    parser.add_argument(
        "--http_proxy",
        required=False,
        help="http proxy address, if required",
    )
    parser.add_argument(
        "--https_proxy",
        required=False,
        help="https proxy address, if required",
    )

    args = parser.parse_args()
    proxies: dict[str, str] = {}
    if args.clear_cache:
        cache.clear()
    if args.http_proxy:
        proxies["http"] = args.http_proxy
    if args.https_proxy:
        proxies["https"] = args.https_proxy
    if args.parser_name == "OPENTARGETS_TARGET":
        extract_open_targets(args.output_path, proxies)



import logging
from collections import defaultdict
from enum import Enum, auto
from typing import Iterable, Optional

import pandas as pd
from kazu.data import OntologyStringResource, MentionConfidence
from kazu.krt.resource_manager import ResourceManager
from kazu.krt.utils import (
    create_new_resource_with_updated_synonyms,
    resource_to_df,
)
from kazu.ontology_preprocessing.curation_utils import OntologyStringConflictAnalyser


class CaseConflictResolutionRequest(Enum):
    PESSIMISTIC = auto()
    OPTIMISTIC = auto()
    CUSTOM = auto()


class ResourceConflict:
    """This class represents a conflict in resources.

    It provides methods to resolve these conflicts optimistically or pessimistically.
    """

    def __init__(self, conflict_dict: dict[str, Iterable[OntologyStringResource]]):
        """
        :param conflict_dict: A dictionary mapping parser names to a list of resources that are in conflict.
        """
        self.parser_to_resource_to_resolution: defaultdict[
            str, dict[OntologyStringResource, Optional[OntologyStringResource]]
        ] = defaultdict(lambda: dict())
        self.parser_names = set()
        self.forms_to_parser = defaultdict(set)
        self.string_set = set()
        self.confidences = set()
        self.cs = set()
        for parser_name, resources in conflict_dict.items():
            self.parser_names.add(parser_name)
            for resource in resources:
                self.parser_to_resource_to_resolution[parser_name][resource] = None
                for syn in resource.active_ner_synonyms():
                    self.forms_to_parser[syn].add(parser_name)
                    self.string_set.add(syn.text)
                    self.confidences.add(syn.mention_confidence)
                    self.cs.add(syn.case_sensitive)

    def _shortest_string_len(self) -> int:
        """Get the length of the shortest string in the conflict.

        :return: The length of the shortest string.
        """
        return min(len(x) for x in self.string_set)

    def __lt__(self, other: "ResourceConflict") -> bool:
        """Compare this ResourceConflict with another based on the length of the
        shortest string in the conflict.

        :param other: The other ResourceConflict to compare with.
        :return: True if this ResourceConflict's shortest string is shorter than the
            other's, False otherwise.
        """
        return self._shortest_string_len() < other._shortest_string_len()

    def batch_resolve(self, optimistic: bool) -> None:
        """Resolve the conflict in batch, changing the parameters of all synonyms at
        once to share the same values.

        The optimistic param indicates the minimum case sensitivity and maximum mention
        confidence should be chosen (otherwise vice-versa)

        :param optimistic: Whether to resolve the conflict optimistically or
            pessimistically.
        """

        new_cs = min(self.cs) if optimistic else max(self.cs)
        new_conf = max(self.confidences) if optimistic else min(self.confidences)
        self._resolve(new_cs=new_cs, new_conf=new_conf)

    def _resolve(self, new_cs: bool, new_conf: MentionConfidence) -> None:
        """Resolve the conflict by creating a new resource with updated synonyms.

        :param new_cs: The new case sensitivity to use.
        :param new_conf: The new mention confidence to use.
        """
        for parser_name, resolution_dict in self.parser_to_resource_to_resolution.items():
            for resource in resolution_dict.keys():
                new_resource = create_new_resource_with_updated_synonyms(new_conf, new_cs, resource)
                self.parser_to_resource_to_resolution[parser_name][resource] = new_resource


class ResourceConflictManager:
    """This class is responsible for managing :class:`.ResourceConflict`\\s in
    resources.

    It provides methods to find conflicts in resources, sync resources for resolved
    string conflicts and find new conflicts.
    """

    def __init__(
        self,
        manager: ResourceManager,
    ):
        """Initialize with a :class:`.ResourceManager`\\.

        :param manager: The :class:`.ResourceManager` to use.
        """
        self.manager = manager
        self.unresolved_conflicts: dict[int, ResourceConflict] = {}
        self.unresolved_conflicts_by_parser: dict[
            tuple[str, OntologyStringResource], ResourceConflict
        ] = {}
        self._init_conflict_maps()

    def sync_resources_for_resolved_resource_conflict_and_find_new_conflicts(
        self, conflict: ResourceConflict
    ) -> None:
        """Sync resources for a resolved conflict and find new conflicts.

        This will refresh the internal map of conflicts.

        :param conflict: a resolved :class:`.ResourceConflict`\\.
        """
        self._sync_resources(conflict)
        self._check_resolved_conflict_for_new_conflicts(conflict)

    def _init_conflict_maps(self) -> None:
        for i, conflict in enumerate(
            self._find_conflicts_in_resources(self.manager.resource_to_parsers.keys())
        ):
            self.unresolved_conflicts[i] = conflict
            self._update_conflict_parser_map((conflict,))

    def _update_conflict_parser_map(self, unresolved_conflicts: Iterable[ResourceConflict]) -> None:
        """Update the conflict parser map with unresolved conflicts.

        :param unresolved_conflicts: The unresolved conflicts to add to the map.
        """
        for conflict in unresolved_conflicts:
            for parser_name, resolution_dict in conflict.parser_to_resource_to_resolution.items():
                for resource in resolution_dict.keys():
                    self.unresolved_conflicts_by_parser[(parser_name, resource)] = conflict

    def _find_conflicts_in_resources(
        self, resources: Iterable[OntologyStringResource]
    ) -> set[ResourceConflict]:
        """Find conflicts in resources.

        :param resources: The resources to check for conflicts against other resources
            currently loaded in the internal resource manager.
        :return: A set of ResourceConflicts.
        """
        logging.info("Looking for conflicts across all parsers...")
        (
            case_conflicts,
            _,
        ) = OntologyStringConflictAnalyser.check_for_case_conflicts_across_resources(
            resources, strict=True  # type: ignore[arg-type]  # dict_keys isn't a subtype of builtin set
        )
        unresolved_conflicts = set()
        for conflict_set in case_conflicts:
            conflict_dict = defaultdict(set)
            for conflict_resource in conflict_set:
                parser_name: str
                for parser_name in self.manager.resource_to_parsers[conflict_resource]:
                    conflict_dict[parser_name].add(conflict_resource)

            conflict = ResourceConflict(dict(conflict_dict))
            unresolved_conflicts.add(conflict)
        return unresolved_conflicts

    def _sync_resources(
        self,
        conflict: ResourceConflict,
    ) -> None:
        """When synching resources, we are implicitly creating a new human curation, as
        the autogenerated set should never be manually changed.

        Therefore, we only care about
        :param conflict:
        :return:
        """
        for parser_name, resolution_map in conflict.parser_to_resource_to_resolution.items():
            for original_resource, new_resource in resolution_map.items():
                assert new_resource is not None
                # pop the key as is no longer valid
                try:
                    self.unresolved_conflicts_by_parser.pop(
                        (
                            parser_name,
                            original_resource,
                        )
                    )
                except KeyError:
                    pass

                self.manager.sync_resources(
                    original_resource=original_resource,
                    new_resource=new_resource,
                    parser_name=parser_name,
                )

    def _check_resolved_conflict_for_new_conflicts(self, conflict: ResourceConflict) -> None:
        parser_name: str
        for parser_name, resolution_map in conflict.parser_to_resource_to_resolution.items():
            for new_resource in resolution_map.values():
                assert new_resource is not None
                new_conflicts = self._find_conflicts_in_resources([new_resource])
                if new_conflicts:
                    self._update_conflict_parser_map(new_conflicts)
        self.unresolved_conflicts = {
            i: conflict
            for i, conflict in enumerate(set(self.unresolved_conflicts_by_parser.values()))
        }

    def _resource_conflict_to_df(self, conflict: ResourceConflict) -> pd.DataFrame:
        data = []
        for parser_name, resource_dict in conflict.parser_to_resource_to_resolution.items():
            df = pd.concat(resource_to_df(resource) for resource in resource_dict.keys())
            df["parser_name"] = parser_name
            data.append(df)
        if len(data) == 1:
            return data[0]
        else:
            return pd.concat(data)

    def summary_df(self) -> pd.DataFrame:
        """Create a summary :class:`~pandas.DataFrame` of unresolved conflicts.

        :return: A DataFrame summarizing the unresolved conflicts.
        """
        data = []
        for i, conflict in self.unresolved_conflicts.items():
            data.append(
                {
                    "parsers": "|".join(sorted(list(conflict.parser_names))),
                    "strings": "|".join(sorted(list(conflict.string_set))),
                    "string len": max(len(x) for x in conflict.string_set),
                    "id": i,
                }
            )
        return pd.DataFrame(data)



"""Some functions to update public resources in the kazu model pack."""
import abc
import datetime
import functools
import logging
import os
import re
import sqlite3
import subprocess
from pathlib import Path
from typing import Optional, cast

import pandas as pd
import rdflib
import requests
from rdflib.query import ResultRow

from kazu.ontology_preprocessing.constants import IDX, DEFAULT_LABEL, MAPPING_TYPE, SYN

logger = logging.getLogger(__name__)


def _get_proxy_args_for_wget() -> list[str]:
    http_proxy = os.getenv("HTTP_PROXY")
    https_proxy = os.getenv("HTTPS_PROXY")
    args = ["wget"]
    proxy_args = []
    if http_proxy:
        proxy_args.append(f"-e http_proxy={http_proxy}")
    if https_proxy:
        proxy_args.append(f"-e https_proxy={https_proxy}")

    if proxy_args:
        args.append("-e use_proxy=yes")
        args.extend(proxy_args)

    return args


@functools.cache
def _cached_request(url: str, local_path: Path) -> None:
    logger.info(f"downloading data from {url}")
    response = requests.get(url, stream=True)
    if response.status_code > 300:
        logger.info("request failed: %s", url)
        logger.info(response)
    else:
        with local_path.open(mode="wb") as f:
            for chunk in response.iter_content(chunk_size=None):
                f.write(chunk)


class OntologyDownloader(abc.ABC):
    @abc.abstractmethod
    def download(self, local_path: Path, skip_download: bool = False) -> Path:
        """Download the ontology to the local path.

        :param local_path: the path to download the ontology to
        :param skip_download: whether to skip the download
        :return: the path to the downloaded ontology
        """
        pass

    def delete_previous(self, local_path: Path) -> None:
        """Delete the previous version of the ontology.

        :param local_path: the path to the ontology to delete
        """
        if local_path.exists():
            os.remove(local_path)

    @abc.abstractmethod
    def version(self, local_path: Optional[Path] = None) -> str:
        """Get the version of the ontology.

        Note that this method should be idempotent, i.e. it should not change the state
        of the ontology. Also, it may be able to determine the version of the ontology
        without querying it directly (e.g. by looking at the file name, or if it is
        known a priori). If this is not the case, you can implement a method here to do
        something more sophisticated, such as querying the ontology directly via sparql.

        Alternatively, it may only be known prior to calling the download method. In
        this case, implementations should store the version as a field, so it can be
        returned from there.

        :param local_path: the path to the ontology
        :return: the version of the ontology
        """
        pass


class SimpleOntologyDownloader(OntologyDownloader):
    def __init__(self, url: str):
        self.url = url

    def download(self, local_path: Path, skip_download: bool = False) -> Path:
        if not skip_download:
            _cached_request(self.url, local_path)
        return local_path

    def version(self, local_path: Optional[Path] = None) -> str:
        return datetime.datetime.now().isoformat()


class OBOOntologyDownloader(SimpleOntologyDownloader):
    def version(self, local_path: Optional[Path] = None) -> str:
        assert local_path is not None
        matcher = re.compile(r"data-version: (.*)$")
        for line in local_path.open(mode="r"):
            match = matcher.match(line)
            if match:
                return match.group(1)
        logger.warning("could not determine version for %s", local_path)
        return super().version(local_path)


class OwlOntologyDownloader(SimpleOntologyDownloader):
    """Used when version info is contained within the ontology."""

    _VERSION_INFO_QUERY = """SELECT ?aname
                    WHERE {
                        ?a rdf:type owl:Ontology .
                        ?a owl:versionInfo ?aname .
                    }"""

    _VERSION_IRI_QUERY = """SELECT ?aname
                    WHERE {
                        ?a rdf:type owl:Ontology .
                        ?a owl:versionIRI ?aname .
                    }"""

    def version(self, local_path: Optional[Path] = None) -> str:
        """Queries the Ontology for owl:versionInfo. Failing that, queries for
        owl:versionIRI. If it can't find it, it falls back to the superclass
        implementation.

        :param local_path:
        :return:
        """
        graph = rdflib.Graph().parse(local_path)
        v_info_result: list[ResultRow] = cast(
            list[ResultRow], list(graph.query(self._VERSION_INFO_QUERY))
        )
        if len(v_info_result) == 1:
            return str(v_info_result[0][0])
        else:
            v_iri_result: list[ResultRow] = cast(
                list[ResultRow], list(graph.query(self._VERSION_IRI_QUERY))
            )
            if len(v_iri_result) == 1:
                return str(v_iri_result[0][0])
            else:
                logger.warning("could not determine versionInfo for %s", local_path)
            return super().version()


class ChemblParquetOntologyDownloader(OntologyDownloader):
    """Downloads the ChEMBL database and exports a subset of it as a parquet file."""

    CHEMBL_TEMPLATE = "https://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_%s/chembl_%s_sqlite.tar.gz"

    CHEMBL_QUERY = f"""SELECT DISTINCT * FROM (
                        SELECT chembl_id AS {IDX}, pref_name AS {DEFAULT_LABEL}, synonyms AS {SYN}, syn_type AS {MAPPING_TYPE}
                        FROM molecule_dictionary AS md
                            JOIN molecule_synonyms ms ON md.molregno = ms.molregno
                        UNION ALL
                        SELECT chembl_id AS {IDX}, pref_name AS {DEFAULT_LABEL}, pref_name AS {SYN}, 'pref_name' AS {MAPPING_TYPE}
                        FROM molecule_dictionary) as t1
                    WHERE t1.DEFAULT_LABEL is not null;
                    """

    CHEMBL_FILENAME_TEMPLATE = "chembl_%s_subset.parquet"

    def __init__(self, chembl_version: str):
        self.chembl_version = chembl_version

    def download(self, local_path: Path, skip_download: bool = False) -> Path:
        chembl_url = self.CHEMBL_TEMPLATE % (self.chembl_version, self.chembl_version)
        # since cheml comes as a zip, we use the parent path to extract it
        parent_path = local_path.parent
        chembl_zip_path = parent_path.joinpath(chembl_url.split("/")[-1])
        if chembl_zip_path.exists() or skip_download:
            logger.info(
                "chembl db already exists %s or skip_download specified, skipping download",
                chembl_zip_path,
            )
        else:

            logger.info("downloading chembl %s", self.chembl_version)

            args = _get_proxy_args_for_wget()
            args.append(chembl_url)
            subprocess.run(
                args,
                cwd=parent_path,
            )

        chembl_unzip_path = parent_path.joinpath(f"chembl_{self.chembl_version}")
        chembl_db_path = chembl_unzip_path.joinpath(
            f"chembl_{self.chembl_version}_sqlite"
        ).joinpath(f"chembl_{self.chembl_version}.db")
        if chembl_db_path.exists():
            logger.info("chembl db already exists %s, skipping extraction", chembl_db_path)
        else:
            logger.info("extracting chembl DB")
            subprocess.run(["tar", "-xvzf", str(chembl_zip_path.absolute())], cwd=parent_path)

        out_path = parent_path.joinpath(self.CHEMBL_FILENAME_TEMPLATE % self.chembl_version)
        df = pd.read_sql(self.CHEMBL_QUERY, sqlite3.connect(chembl_db_path))
        df.to_parquet(out_path)

        return out_path

    def version(self, local_path: Optional[Path] = None) -> str:
        return self.chembl_version


class OpenTargetsOntologyDownloader(OntologyDownloader):
    OT_PREFIX = "ftp://ftp.ebi.ac.uk/pub/databases/opentargets/platform/"

    def __init__(self, open_targets_version: str, open_targets_dataset_name: str):
        self.open_targets_dataset_name = open_targets_dataset_name
        self.open_targets_version = open_targets_version

    @staticmethod
    @functools.cache
    def _cached_wget(full_url: str, cwd: Path) -> None:
        args = _get_proxy_args_for_wget()
        args.extend(
            [
                "--recursive",
                "--no-parent",
                "--no-host-directories",
                "--cut-dirs",
                "8",
                full_url,
            ]
        )
        subprocess.run(args, cwd=cwd)

    def download(self, local_path: Path, skip_download: bool = False) -> Path:
        if not skip_download:
            full_url = f"{self.OT_PREFIX}{self.open_targets_version}/output/etl/json/{self.open_targets_dataset_name}"
            OpenTargetsOntologyDownloader._cached_wget(full_url, local_path.parent)
        return local_path

    def version(self, local_path: Optional[Path] = None) -> str:
        return self.open_targets_version

    def delete_previous(self, local_path: Path) -> None:
        """We could use rmtree here but it's safer to just remove the files we know we
        downloaded.

        :param local_path:
        :return:
        """
        if local_path.exists():
            for file in local_path.iterdir():
                if file.name.startswith("part-") or file.name.startswith("_SUCCESS"):
                    os.remove(file)
            os.removedirs(local_path)



import functools
import inspect
from pathlib import Path
from typing import Any, Optional, cast

import pandas as pd
import streamlit as st
from hydra.utils import instantiate
from omegaconf import DictConfig, open_dict, OmegaConf

from kazu.data import OntologyStringResource
from kazu.krt.utils import resource_to_df
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.ontology_preprocessing.curation_utils import dump_ontology_string_resources
from kazu.ontology_preprocessing.ontology_upgrade_report import OntologyUpgradeReport
from kazu.krt import load_config


class OntologyUpdateManager:
    """A class to manage the updating the raw ontology files of ontology parsers."""

    def __init__(self, parser_name: str):
        self.original_parser_cfg: DictConfig = self.name_to_parser_cfg()[parser_name]
        self.parser = instantiate(self.original_parser_cfg)
        self.new_parser_cfg: DictConfig = self.original_parser_cfg.copy()

    def get_downloader_args_and_types(self) -> dict[str, type]:
        """Get the types of the arguments of the
        :class:`kazu.ontology_preprocessing.downloads.OntologyDownloader` class."""
        return inspect.getfullargspec(
            self.parser.ontology_downloader.__class__.__init__
        ).annotations

    def get_downloader_args_and_vals(self) -> dict[str, Any]:
        """Get the arguments and their values of the configured
        :class:`kazu.ontology_preprocessing.downloads.OntologyDownloader` class."""
        configured_args_dict = dict(self.original_parser_cfg.ontology_downloader)
        # remove target as it is not an argument
        configured_args_dict.pop("_target_")
        return configured_args_dict

    def display_new_parser_config_as_yaml(self) -> str:
        return OmegaConf.to_yaml(self.new_parser_cfg)

    def display_new_downloader_config_as_yaml(self) -> str:
        return OmegaConf.to_yaml(self.new_parser_cfg.ontology_downloader)

    @staticmethod
    def name_to_parser_cfg() -> dict[str, DictConfig]:
        """Get the names of the parsers and their configurations.

        :return:
        """
        result = {}
        cfg = load_config()
        for parser_key, parser_config_dict in cfg.ontologies.parsers.items():
            result[parser_config_dict.name] = parser_config_dict
        return result

    def modify_downloader_cfg_with_new_args(self, kwargs: dict[str, Any]) -> None:
        """Modify the downloader configuration with the new arguments.

        :param kwargs:
        :return:
        """
        with open_dict(self.new_parser_cfg):
            for arg_name, arg_val in kwargs.items():
                self.new_parser_cfg.ontology_downloader[arg_name] = arg_val

    def modify_parser_cfg_with_new_args(self) -> None:
        """Modify the parser configuration with the new downloader configuration.

        :return:
        """
        parser = instantiate(self.new_parser_cfg)
        new_in_path = parser.ontology_downloader.download(parser.in_path, skip_download=True)
        with open_dict(self.new_parser_cfg):
            self.new_parser_cfg.data_origin = parser.ontology_downloader.version(parser.in_path)
            self.new_parser_cfg.in_path = str(new_in_path.absolute())

    def instantiate_parser_with_new_config(self) -> OntologyParser:
        """Instantiate a new parser with the new configuration.

        :return:
        """
        return cast(OntologyParser, instantiate(self.new_parser_cfg))

    @functools.cache
    def get_or_build_upgrade_report(self) -> OntologyUpgradeReport:
        """Get the upgrade report if it exists, otherwise build it.

        :return:
        """

        parser = self.instantiate_parser_with_new_config()
        return parser.upgrade_ontology_version()

    def download_ontology_with_new_config(self) -> None:
        """Download the ontology with the new configuration.

        :return:
        """
        parser = self.instantiate_parser_with_new_config()
        assert parser.ontology_downloader is not None
        parser.download_ontology()

    def obsolete_df(self) -> Optional[pd.DataFrame]:
        """Get the dataframe of the obsolete resources.

        :return:
        """
        return self._df_from_set(
            self.get_or_build_upgrade_report().obsolete_resources_after_upgrade
        )

    def _df_from_set(self, resources: set[OntologyStringResource]) -> Optional[pd.DataFrame]:
        if resources:
            df = pd.concat(resource_to_df(resource, True) for resource in resources)
            df["string_length"] = df["text"].apply(len)
            return df
        else:
            return None

    def novel_df(self) -> Optional[pd.DataFrame]:
        """Get the dataframe of the novel resources.

        :return:
        """
        return self._df_from_set(self.get_or_build_upgrade_report().new_resources_after_upgrade)

    def write_new_defaults_to_model_pack(self) -> Path:
        """Write the new default resources to the model pack.

        :return:
        """
        defaults_path = cast(Path, self.parser.ontology_auto_generated_resources_set_path)
        with st.spinner(f"updating defaults at {defaults_path}"):
            dump_ontology_string_resources(
                self.get_or_build_upgrade_report().new_version_auto_generated_resources_clean,
                defaults_path,
                force=True,
            )
        return defaults_path



import streamlit as st

from kazu.krt.components import ParserSelector
from kazu.krt.ontology_update_editor.utils import OntologyUpdateManager
from kazu.krt.utils import get_resource_manager


@st.cache_resource
def get_upgrade_manager(parser_name: str) -> OntologyUpdateManager:
    """Get an OntologyUpdateManager instance for the given parser_name.

    :param parser_name:
    :return:
    """
    return OntologyUpdateManager(parser_name)


class OntologyUpdateForm:
    DOWNLOAD_COMPLETED_OR_SKIPPED = "download_completed"
    ATTEMPT_DOWNLOAD = "download_skipped"
    UPDATES_SAVED = "updates_saved"

    @staticmethod
    def get_download_completed_or_skipped(parser_name: str) -> bool:
        return bool(
            st.session_state.get(
                f"{parser_name}_{OntologyUpdateForm.DOWNLOAD_COMPLETED_OR_SKIPPED}"
            )
        )

    @staticmethod
    def set_download_completed_or_skipped(parser_name: str) -> None:
        st.session_state[f"{parser_name}_{OntologyUpdateForm.DOWNLOAD_COMPLETED_OR_SKIPPED}"] = True

    @staticmethod
    def get_updates_saved_key(parser_name: str) -> str:
        return f"{parser_name}_{OntologyUpdateForm.UPDATES_SAVED}"

    @staticmethod
    def get_attempt_download_key(parser_name: str) -> str:
        return f"{parser_name}_{OntologyUpdateForm.ATTEMPT_DOWNLOAD}"

    @staticmethod
    def get_updates_saved(parser_name: str) -> bool:
        return bool(st.session_state.get(OntologyUpdateForm.get_updates_saved_key(parser_name)))

    @staticmethod
    def display_download_config_modifier_form(manager: OntologyUpdateManager) -> None:
        """

        :param manager:
        :return:
        """
        with st.form("config_mod_form"):
            for arg, val in manager.get_downloader_args_and_vals().items():
                st.text_input(arg, value=val, key=arg)
            st.form_submit_button(
                "Submit",
                on_click=OntologyUpdateForm._submit_download_config_modifier_form,
                args=(manager,),
            )

    @staticmethod
    def _submit_download_config_modifier_form(manager: OntologyUpdateManager) -> None:
        updated_args = {}
        for arg, arg_type in manager.get_downloader_args_and_types().items():
            updated_args[arg] = arg_type(st.session_state[arg])
        manager.modify_downloader_cfg_with_new_args(updated_args)
        manager.modify_parser_cfg_with_new_args()

    @staticmethod
    def display_download_form(manager: OntologyUpdateManager) -> None:
        if not OntologyUpdateForm.get_download_completed_or_skipped(manager.parser.name):
            with st.form("downloader_config_form"):
                st.radio(
                    "Download resource",
                    options=["yes", "no"],
                    index=0,
                    key=OntologyUpdateForm.get_attempt_download_key(manager.parser.name),
                )
                st.form_submit_button(
                    "Continue - warning: downloading may delete the original file/folder in your model pack",
                    on_click=OntologyUpdateForm._submit_download_form,
                    args=(manager,),
                )

    @staticmethod
    def _submit_download_form(manager: OntologyUpdateManager) -> None:
        if (
            st.session_state[OntologyUpdateForm.get_attempt_download_key(manager.parser.name)]
            == "yes"
        ):
            with st.spinner("downloading ontology"):
                manager.download_ontology_with_new_config()
        OntologyUpdateForm.set_download_completed_or_skipped(manager.parser.name)

    @staticmethod
    def _write_new_defaults_to_model_pack(manager: OntologyUpdateManager) -> None:
        with st.spinner("writing new defaults to model pack"):
            manager.write_new_defaults_to_model_pack()
        # writing defaults to model pack will invalidate the resource manager cache
        get_resource_manager.clear()  # type: ignore[attr-defined]

    @staticmethod
    def display_main_form() -> None:
        ParserSelector.display_parser_selector(clear_state=True)
        parser_name = ParserSelector.get_selected_parser_name()
        if parser_name:
            with st.spinner(f"preparing to update {parser_name}"):
                manager = get_upgrade_manager(parser_name)
            if not manager.parser.ontology_downloader:
                st.write(
                    "no downloader available for this parser. To update, manually put the resource in the model pack and delete the default resources folder."
                )
            else:
                if not OntologyUpdateForm.get_download_completed_or_skipped(parser_name):
                    st.header("set the parameters for the download")
                    OntologyUpdateForm.display_download_config_modifier_form(manager)
                    st.markdown("### new downloader yaml config is as follows:")
                    st.markdown(
                        f"""```yaml
                    {manager.display_new_downloader_config_as_yaml()}```"""
                    )
                    OntologyUpdateForm.display_download_form(manager)
                if OntologyUpdateForm.get_download_completed_or_skipped(parser_name):
                    manager.modify_parser_cfg_with_new_args()
                    st.markdown(
                        "### new parser config (note, you may need to correct interpolated paths):"
                    )
                    st.markdown(
                        f"""```yaml
                    {manager.display_new_parser_config_as_yaml()}```"""
                    )

                    if not OntologyUpdateForm.get_updates_saved(parser_name):
                        with st.spinner("building upgrade_report"):

                            report = manager.get_or_build_upgrade_report()

                        st.write(
                            f"found {len(report.obsolete_resources_after_upgrade)} obsolete resources"
                        )
                        st.write(manager.obsolete_df())
                        st.write(f"found {len(report.new_resources_after_upgrade)} new resources")
                        st.write(manager.novel_df())
                        st.button(
                            f"click here to update the default resources for {parser_name}",
                            on_click=OntologyUpdateForm._write_new_defaults_to_model_pack,
                            args=(manager,),
                            key=OntologyUpdateForm.get_updates_saved_key(parser_name),
                        )
                    else:
                        st.write("updates saved")



import dataclasses
from typing import Iterable, Protocol

from kazu.data import OntologyStringResource, MentionConfidence, OntologyStringBehaviour
from kazu.utils.string_normalizer import StringNormalizer


class AutoCurationAction(Protocol):
    def __call__(self, resource: OntologyStringResource) -> OntologyStringResource:
        raise NotImplementedError


class SymbolicToCaseSensitiveAction(AutoCurationAction):
    def __init__(self, entity_class: str):
        self.entity_class = entity_class

    def __call__(self, resource: OntologyStringResource) -> OntologyStringResource:
        all_symbolic = all(
            StringNormalizer.classify_symbolic(syn.text, entity_class=self.entity_class)
            for syn in resource.original_synonyms
        )
        if all_symbolic:
            return dataclasses.replace(
                resource,
                original_synonyms=frozenset(
                    dataclasses.replace(syn, case_sensitive=True)
                    for syn in resource.original_synonyms
                ),
                alternative_synonyms=frozenset(
                    dataclasses.replace(syn, case_sensitive=True)
                    for syn in resource.alternative_synonyms
                ),
            )
        else:
            return resource


class IsCommmonWord(AutoCurationAction):
    def __init__(self, path: str):
        with open(path, mode="r") as inf:
            self.common_words = {line.rstrip() for line in inf.readlines()}

    def __call__(self, resource: OntologyStringResource) -> OntologyStringResource:
        found_common = (
            all(word in self.common_words for word in syn.text.lower().split())
            for syn in resource.original_synonyms
        )

        if any(found_common):
            return dataclasses.replace(
                resource,
                original_synonyms=frozenset(
                    dataclasses.replace(syn, mention_confidence=MentionConfidence.POSSIBLE)
                    for syn in resource.original_synonyms
                ),
                alternative_synonyms=frozenset(
                    dataclasses.replace(syn, mention_confidence=MentionConfidence.POSSIBLE)
                    for syn in resource.alternative_synonyms
                ),
            )
        else:
            return resource


class MinLength(AutoCurationAction):
    def __init__(self, min_len: int = 2):
        self.min_len = min_len

    def __call__(self, resource: OntologyStringResource) -> OntologyStringResource:
        for syn in resource.original_synonyms:
            if len(syn.text) < self.min_len:
                return dataclasses.replace(
                    resource, behaviour=OntologyStringBehaviour.DROP_FOR_LINKING
                )

        return resource


class MaxLength(AutoCurationAction):
    """Drop resources that exceed a maximum string length."""

    def __init__(self, max_len: int = 60):
        self.max_len = max_len

    def __call__(self, resource: OntologyStringResource) -> OntologyStringResource:
        if any(len(syn.text) > self.max_len for syn in resource.original_synonyms):
            return dataclasses.replace(resource, behaviour=OntologyStringBehaviour.DROP_FOR_LINKING)
        else:
            return resource


def is_upper_case_word_to_case_insensitive(
    resource: OntologyStringResource,
) -> OntologyStringResource:
    """Make Resources where all original synonyms are all uppercase alphabetical
    characters case-insensitive.

    Some data sources use all-caps strings for nouns that can be considered case-
    insensitive (e.g. Chembl).

    :param resource:
    :return:
    """

    if all(syn.text.isupper() and syn.text.isalpha() for syn in resource.original_synonyms):
        return dataclasses.replace(
            resource,
            original_synonyms=frozenset(
                dataclasses.replace(syn, case_sensitive=False) for syn in resource.original_synonyms
            ),
            alternative_synonyms=frozenset(
                dataclasses.replace(syn, case_sensitive=False)
                for syn in resource.alternative_synonyms
            ),
        )
    else:
        return resource


def initial_lowercase_then_upper_to_case_sensitive(
    resource: OntologyStringResource,
) -> OntologyStringResource:
    """If a synonym starts with a lowercase character followed by an uppercase
    character, then all synonyms should be case-sensitive.

    E.g. "eGFR" vs "EGFR".

    :param resource:
    :return:
    """
    if any(
        len(syn.text) >= 2 and syn.text[0].islower() and syn.text[1].isupper()
        for syn in resource.original_synonyms
    ):
        return dataclasses.replace(
            resource,
            original_synonyms=frozenset(
                dataclasses.replace(syn, case_sensitive=True) for syn in resource.original_synonyms
            ),
            alternative_synonyms=frozenset(
                dataclasses.replace(syn, case_sensitive=True)
                for syn in resource.alternative_synonyms
            ),
        )
    else:
        return resource


class LikelyAcronym(AutoCurationAction):
    """If all synonyms are less than or equal to the specified length, and are all upper
    case, give a confidence of POSSIBLE to all forms."""

    def __init__(self, max_len_to_consider: int = 5):
        self.max_len_to_consider = max_len_to_consider

    def __call__(self, resource: OntologyStringResource) -> OntologyStringResource:

        if all(
            len(syn.text) <= self.max_len_to_consider and syn.text.isupper()
            for syn in resource.original_synonyms
        ):
            return dataclasses.replace(
                resource,
                original_synonyms=frozenset(
                    dataclasses.replace(syn, mention_confidence=MentionConfidence.POSSIBLE)
                    for syn in resource.original_synonyms
                ),
                alternative_synonyms=frozenset(
                    dataclasses.replace(syn, mention_confidence=MentionConfidence.POSSIBLE)
                    for syn in resource.alternative_synonyms
                ),
            )
        else:
            return resource


class AutoCurator:
    def __init__(self, actions: list[AutoCurationAction]):
        self.actions = actions

    def __call__(self, resources: set[OntologyStringResource]) -> Iterable[OntologyStringResource]:
        for resource in resources:
            for action in self.actions:
                resource = action(resource)

            yield resource



import logging
from pathlib import Path

from kazu.data import (
    OntologyStringResource,
)
from kazu.ontology_preprocessing.curation_utils import (
    dump_ontology_string_resources,
)

logger = logging.getLogger(__name__)

_ONTOLOGY_UPGRADE_REPORT_DIR = "_ontology_upgrade_report"


class OntologyUpgradeReport:
    """A report on the delta of :class:`~.OntologyStringResource`\\s generated between
    two versions of an ontology, or the same version of the ontology with a different
    configuration of :class:`~.AutoCurator`\\.

    This is useful to highlight which resources are novel or obsolete between versions.
    """

    def __init__(
        self,
        new_version_auto_generated_resources_clean: set[OntologyStringResource],
        previous_version_auto_generated_resources_clean: set[OntologyStringResource],
    ):
        """

        :param new_version_auto_generated_resources_clean: resources from new version.
        :param previous_version_auto_generated_resources_clean: resources from previous version.
        """
        self.new_version_auto_generated_resources_clean = new_version_auto_generated_resources_clean
        self.previous_version_auto_generated_resources_clean = (
            previous_version_auto_generated_resources_clean
        )

    @property
    def new_resources_after_upgrade(self) -> set[OntologyStringResource]:
        """Novel resources generated after upgrade/config change."""
        return self.new_version_auto_generated_resources_clean.difference(
            self.previous_version_auto_generated_resources_clean
        )

    @property
    def obsolete_resources_after_upgrade(self) -> set[OntologyStringResource]:
        """Resources that are now obsolete (i.e. no longer do anything) after upgrade
        :return:"""
        return self.previous_version_auto_generated_resources_clean.difference(
            self.new_version_auto_generated_resources_clean
        )

    def write_report(self, ontology_data_path: Path, parser_name: str) -> None:
        """Write a report for human review of the delta.

        :param ontology_data_path:
        :param parser_name:
        :return:
        """

        upgrade_report_path = ontology_data_path.parent.joinpath(_ONTOLOGY_UPGRADE_REPORT_DIR)
        logger.info(
            "%s writing novel/obsolete resource reports after upgrade (to %s)",
            parser_name,
            upgrade_report_path,
        )

        maybe_new_resources = self.new_resources_after_upgrade
        if maybe_new_resources:
            dump_ontology_string_resources(
                maybe_new_resources,
                upgrade_report_path.joinpath("novel_auto_generated_resources.jsonl"),
                force=True,
            )
        maybe_obsolete_resources = self.obsolete_resources_after_upgrade
        if maybe_obsolete_resources:
            dump_ontology_string_resources(
                maybe_obsolete_resources,
                upgrade_report_path.joinpath("obsolete_auto_generated_resources.jsonl"),
                force=True,
            )



import dataclasses
import json
import logging
import shutil
from collections import defaultdict, Counter
from collections.abc import Iterable
from enum import auto
from typing import Optional, Literal, Any

from kazu.data import (
    EquivalentIdSet,
    EquivalentIdAggregationStrategy,
    LinkingCandidate,
    OntologyStringResource,
    ParserBehaviour,
    OntologyStringBehaviour,
    AssociatedIdSets,
    GlobalParserActions,
    AutoNameEnum,
    Synonym,
    MentionConfidence,
)
from kazu.database.in_memory_db import (
    NormalisedSynonymStr,
    Idx,
)
from kazu.utils.string_normalizer import StringNormalizer
from kazu.utils.utils import (
    PathLike,
    as_path,
)

logger = logging.getLogger(__name__)


class CurationError(Exception):
    pass


def load_ontology_string_resources(
    path: PathLike,
) -> set[OntologyStringResource]:
    """Load :class:`kazu.data.OntologyStringResource`\\ s from a file path or directory.

    :param path: path to a jsonl file or directory of jsonl files that map to :class:`kazu.data.OntologyStringResource`
    :return:
    """
    resources_path = as_path(path)
    resources: set[OntologyStringResource] = set()
    if not resources_path.exists():
        raise ValueError(f"resources file does not exist at: {path}")

    files = resources_path.iterdir() if resources_path.is_dir() else (resources_path,)
    for f in files:
        with f.open(mode="r") as jsonlf:
            resources.update(OntologyStringResource.from_json(line) for line in jsonlf)

    return resources


def _resource_sort_reduce(resource: OntologyStringResource) -> tuple[int, str]:
    return min((len(syn.text), syn.text) for syn in resource.original_synonyms)


def batch(
    iterable: Iterable[OntologyStringResource], n: int = 1
) -> Iterable[list[OntologyStringResource]]:
    lst = sorted(iterable, key=_resource_sort_reduce)
    length = len(lst)
    for ndx in range(0, length, n):
        yield lst[ndx : ndx + n]


def dump_ontology_string_resources(
    resources: Iterable[OntologyStringResource],
    path: PathLike,
    force: bool = False,
    split_at: int = 10000,
) -> None:
    """Dump an iterable of :class:`kazu.data.OntologyStringResource`\\s to the file
    system.

    :param resources: resources to dump
    :param path: path to a directory of json lines files that map to :class:`kazu.data.OntologyStringResource`
    :param force: override existing directory, if it exists
    :param split_at: number of lines per partition
    :return:
    """
    resources_path = as_path(path)
    if resources_path.exists():
        if not resources_path.is_dir():
            raise ValueError(f"file already exists at {path} and is not a directory")
        elif not force:
            raise ValueError(f"directory already exists at {path}")
        else:
            shutil.rmtree(resources_path)

    resources_path.mkdir(parents=True)
    for i, resources_partition in enumerate(batch(resources, n=split_at)):
        with resources_path.joinpath(f"{i}.jsonl").open(mode="w") as jsonlf:
            for resource in resources_partition:
                jsonlf.write(resource.to_json() + "\n")


def load_global_actions(
    path: PathLike,
) -> GlobalParserActions:
    """Load an instance of :class:`.GlobalParserActions` from a file path.

    :param path: path to a json serialised GlobalParserActions
    :return:
    """
    global_actions_path = as_path(path)
    if global_actions_path.exists():
        with global_actions_path.open(mode="r") as jsonf:
            global_actions = GlobalParserActions.from_dict(json.load(jsonf))
    else:
        raise ValueError(f"global actions do not exist at {path}")
    return global_actions


_CURATION_REPORT_FILENAME = "_curation_report"
_ONTOLOGY_MERGE_REPORT_DIR = "_ontology_merge_report"
CLEAN_RESOURCES_FN = "clean_resources.jsonl"
MERGED_RESOURCES_FN = "merged_resources.jsonl"
NORM_CONFLICT_RESOURCES_FN = "normalisation_conflicts.jsonl"
OBSOLETE_RESOURCES_FN = "obsolete_human_curations_set.jsonl"
SUPERFLUOUS_RESOURCES_FN = "superfluous_human_curations_set.jsonl"
EFFECTIVE_RESOURCES_FN = "effective_human_curations.jsonl"


@dataclasses.dataclass
class OntologyResourceSetConflictReport:
    #: Resources with no conflicts
    clean_resources: set[OntologyStringResource]
    #: Resources that can be safely merged without affecting :class:`.OntologyStringBehaviour`. However,
    #: may still conflict on :class:`.MentionConfidence` and/or case sensitivity
    merged_resources: set[OntologyStringResource]
    #: Resources that conflict on normalisation value
    normalisation_conflicts: set[frozenset[OntologyStringResource]]
    #: Resources that conflict on case
    case_conflicts: set[frozenset[OntologyStringResource]]

    def write_normalisation_conflict_report(
        self,
        path: PathLike,
    ) -> None:
        report_path = as_path(path)
        report_path.mkdir(parents=True, exist_ok=True)
        path_to_resources: dict[str, set[OntologyStringResource]] = {
            CLEAN_RESOURCES_FN: self.clean_resources,
            MERGED_RESOURCES_FN: self.merged_resources,
        }
        for path_name, resource_data in path_to_resources.items():
            if resource_data:
                with report_path.joinpath(path_name).open(mode="w") as f:
                    for resource in resource_data:
                        f.write(resource.to_json() + "\n")

        if self.normalisation_conflicts:
            with report_path.joinpath(NORM_CONFLICT_RESOURCES_FN).open(mode="w") as f:
                for resource_set in self.normalisation_conflicts:
                    for resource in resource_set:
                        f.write(resource.to_json() + "\n")


@dataclasses.dataclass
class OntologyResourceSetMergeReport:
    #: human resources that no longer match any strings in the underlying parser data
    obsolete_resources: set[OntologyStringResource]
    #: human and autogenerated resources that are actively in use
    effective_resources: set[OntologyStringResource]
    #: human resources that match any strings in the underlying parser data, but result in the same behaviour
    #: as the autogenerated resource, and therefore can be eliminated to reduce the management burden of human
    #: resources
    superfluous_resources: set[OntologyStringResource]
    #: a tuple of human resource/autogenerated resource that only partially match on their original forms,
    #: suggesting that the underlying parser data has changed in some way since the last upgrade. In this
    #: scenario, it's recommended to replace the human form with the new autogenerated version for consistency
    resources_with_discrepancies: set[tuple[OntologyStringResource, OntologyStringResource]]

    def write_ontology_merge_report(self, path: PathLike) -> None:
        report_path = as_path(path)
        filename_to_resources: dict[str, set[OntologyStringResource]] = {
            OBSOLETE_RESOURCES_FN: self.obsolete_resources,
            SUPERFLUOUS_RESOURCES_FN: self.superfluous_resources,
            EFFECTIVE_RESOURCES_FN: self.effective_resources,
        }
        for filename, resources in filename_to_resources.items():
            if resources:
                dump_ontology_string_resources(
                    resources,
                    report_path.joinpath(filename),
                    force=True,
                )

        if self.resources_with_discrepancies:
            with report_path.joinpath("human_curations_with_synonym_discrepancies_set.jsonl").open(
                mode="w"
            ) as jsonlf:
                for (
                    human_curation,
                    default_resource,
                ) in self.resources_with_discrepancies:
                    jsonlf.write("HUMAN:" + "\n")
                    jsonlf.write(human_curation.to_json() + "\n")
                    jsonlf.write("DEFAULT:" + "\n")
                    jsonlf.write(default_resource.to_json() + "\n")
                    jsonlf.write("\n\n")


@dataclasses.dataclass
class OntologyResourceSetCompleteReport:
    """Describes the state of :class:`kazu.data.OntologyStringResource`\\s configured
    for a given :class:`kazu.ontology_preprocessing.base.OntologyParser`, such as any
    conflicts in how string matching is configured, or discrepancies between resources
    produced by :class:`kazu.ontology_preprocessing.autocuration.AutoCurator` and their
    human curated overrrides."""

    #: parser linking candidates before they are processed with :class:`.OntologyStringResource`\\s
    intermediate_linking_candidates: set[LinkingCandidate]
    #: report of resource conflict resolution after merging of autogenerated resources and human curations
    final_conflict_report: OntologyResourceSetConflictReport
    #: report of resource conflict in human curation set, if available
    human_conflict_report: Optional[OntologyResourceSetConflictReport] = None
    #: report of result of merging human and autogenerated resources, if available
    merge_report: Optional[OntologyResourceSetMergeReport] = None

    def write_reports_for_parser(self, path: PathLike, parser_name: str) -> None:
        curation_report_path = as_path(path).joinpath(f"{parser_name}{_CURATION_REPORT_FILENAME}")
        if curation_report_path.exists():
            shutil.rmtree(curation_report_path)

        curation_report_path.mkdir()

        human_and_auto_generated_resources_conflict_report_path = curation_report_path.joinpath(
            "active_resources_conflict_report"
        )
        human_and_auto_generated_resources_conflict_report_path.mkdir()
        self.final_conflict_report.write_normalisation_conflict_report(
            human_and_auto_generated_resources_conflict_report_path
        )

        if self.human_conflict_report:
            human_curation_set_report_path = curation_report_path.joinpath(
                "human_curation_conflict_report"
            )
            human_curation_set_report_path.mkdir()
            self.human_conflict_report.write_normalisation_conflict_report(
                human_curation_set_report_path
            )

        if self.merge_report:
            logger.info(
                "%s reporting discrepancies in human curation set and auto generated resources",
                parser_name,
            )
            merge_report_path = as_path(path).joinpath(f"{parser_name}{_ONTOLOGY_MERGE_REPORT_DIR}")
            merge_report_path.mkdir()
            self.merge_report.write_ontology_merge_report(merge_report_path)


class AutofixStrategy(AutoNameEnum):
    OPTIMISTIC = auto()
    PESSIMISTIC = auto()
    NONE = auto()


class OntologyStringConflictAnalyser:
    """Find and potentially fix conflicting behaviour in a set of
    :class:`kazu.data.OntologyStringResource`\\s."""

    def __init__(self, entity_class: str, autofix: AutofixStrategy = AutofixStrategy.NONE):
        """

        :param entity_class: entity class that this analyzer will handle
        :param autofix: Should any conflicts be automatically fixed, such that the
            behaviour is consistent within this set? Note that this does not guarantee that
            the optimal behaviour for a conflict is preserved.
        """
        self.entity_class = entity_class
        self.autofix = autofix

    def verify_resource_set_integrity(
        self, resources: set[OntologyStringResource]
    ) -> OntologyResourceSetConflictReport:
        """Verify that a set of resources has consistent behaviour.

        Conflicts can occur for the following reasons:

        1) If two or more resources normalise to the same string,
           but have different :class:`kazu.data.OntologyStringBehaviour`\\.

        2) If two or more resources normalise to the same value,
           but have different associated ID sets specified, such that one would
           override the other.

        3) If two or more resources have conflicting values for case sensitivity and
           :class:`kazu.data.MentionConfidence`\\. E.g. A case-insensitive resource
           cannot have a higher mention confidence value than a case-sensitive one for
           the same string.

        :param resources:
        :return:
        :raises CurationError: if one or more resources produce multiple normalised values
        """

        resources = set(resources)
        (
            merged_resources,
            eliminated_resources,
            normalisation_conflicts,
        ) = self.check_for_normalised_behaviour_conflicts_and_merge_if_possible(resources)

        resources.difference_update(eliminated_resources)
        resources.update(merged_resources)
        # remove the conflicts from the working set
        for conflict_set in normalisation_conflicts:
            resources.difference_update(conflict_set)
        if self.autofix is not AutofixStrategy.NONE:

            logger.info("Resources will be automatically fixed")
            autofixed_resources_from_conflicts = self.autofix_resources(normalisation_conflicts)
            # add the fixed resources to the merged set
            merged_resources.update(autofixed_resources_from_conflicts)
            resources.update(autofixed_resources_from_conflicts)
            case_conflicts, clean_resources = self.check_for_case_conflicts_across_resources(
                resources
            )
            clean_resources.update(self.autofix_resources(case_conflicts))
            normalisation_conflicts = set()
            case_conflicts = set()
        else:
            case_conflicts, clean_resources = self.check_for_case_conflicts_across_resources(
                resources
            )

        return OntologyResourceSetConflictReport(
            clean_resources=clean_resources,
            merged_resources=merged_resources,
            normalisation_conflicts=normalisation_conflicts,
            case_conflicts=case_conflicts,
        )

    def autofix_resources(
        self, resource_conflicts: set[frozenset[OntologyStringResource]]
    ) -> set[OntologyStringResource]:
        """Fix conflicts in resources by producing a new set of resources with
        consistent behaviour.

        This ensures that there are no conflicts regarding any of these properties:

        * the combination of case sensitivity and mention confidence
        * associated id set
        * resource behaviour

        :param resource_conflicts:
        :return:
        """
        cleaned_resources = set()
        agg_func = max if self.autofix is AutofixStrategy.OPTIMISTIC else min
        for conflicted_set in resource_conflicts:
            original_synonyms_by_syn_norm: defaultdict[str, set[Synonym]] = defaultdict(set)
            alt_syns_by_syn_norm: defaultdict[str, set[Synonym]] = defaultdict(set)
            case_sensitivities = set()
            syn_string_lower_to_confidence = defaultdict(set)
            assoc_id_sets: set[EquivalentIdSet] = set()
            behaviours: set[OntologyStringBehaviour] = set()
            for resource in conflicted_set:
                syn_norm = resource.syn_norm_for_linking(self.entity_class)
                original_synonyms_by_syn_norm[syn_norm].update(resource.original_synonyms)
                alt_syns_by_syn_norm[syn_norm].update(resource.alternative_synonyms)
                behaviours.add(resource.behaviour)
                for syn in resource.all_synonyms():
                    syn_string_lower_to_confidence[syn.text.lower()].add(syn.mention_confidence)
                    case_sensitivities.add(syn.case_sensitive)
                if resource.associated_id_sets is not None:
                    assoc_id_sets.update(resource.associated_id_sets)

            if OntologyStringBehaviour.DROP_FOR_LINKING in behaviours:
                chosen_behaviour = OntologyStringBehaviour.DROP_FOR_LINKING
            elif OntologyStringBehaviour.ADD_FOR_LINKING_ONLY in behaviours:
                chosen_behaviour = OntologyStringBehaviour.ADD_FOR_LINKING_ONLY
            else:
                chosen_behaviour = OntologyStringBehaviour.ADD_FOR_NER_AND_LINKING

            chosen_case_sensitivity = (
                min(case_sensitivities)
                if self.autofix is AutofixStrategy.OPTIMISTIC
                else max(case_sensitivities)
            )

            for syn_norm, mergeable_original_synonyms in original_synonyms_by_syn_norm.items():
                cleaned_resources.add(
                    OntologyStringResource(
                        behaviour=chosen_behaviour,
                        original_synonyms=frozenset(
                            dataclasses.replace(
                                syn,
                                case_sensitive=chosen_case_sensitivity,
                                mention_confidence=agg_func(
                                    syn_string_lower_to_confidence[syn.text.lower()]
                                ),
                            )
                            for syn in mergeable_original_synonyms
                        ),
                        alternative_synonyms=frozenset(
                            dataclasses.replace(
                                syn,
                                case_sensitive=chosen_case_sensitivity,
                                mention_confidence=agg_func(
                                    syn_string_lower_to_confidence[syn.text.lower()]
                                ),
                            )
                            for syn in alt_syns_by_syn_norm.get(syn_norm, set())
                        ),
                        associated_id_sets=frozenset(assoc_id_sets)
                        if len(assoc_id_sets) > 0
                        else None,
                    )
                )
        return cleaned_resources

    @staticmethod
    def check_for_case_conflicts_across_resources(
        resources: set[OntologyStringResource], strict: bool = False
    ) -> tuple[set[frozenset[OntologyStringResource]], set[OntologyStringResource]]:
        """Find conflicts in case sensitivity within a set of resources.

        Conflicts can occur when strings differ by case sensitivity, and a
        case-insensitive synonym will produce a :class:`.MentionConfidence`
        of equal or higher rank than a case-sensitive one.

        :param resources:
        :param strict: if True, then the function will return True if there
            are multiple mention confidences for a given string, regardless of case sensitivity
        :return: a set of conflicted subsets, and a set of clean resources.
        """

        maybe_good_resources_by_active_syn_lower = (
            OntologyStringConflictAnalyser.build_synonym_defaultdict(resources)
        )

        return OntologyStringConflictAnalyser.find_case_conflicts(
            maybe_good_resources_by_active_syn_lower, strict=strict
        )

    @staticmethod
    def find_case_conflicts(
        maybe_good_resources_by_active_syn_lower: defaultdict[str, set[OntologyStringResource]],
        strict: bool = False,
    ) -> tuple[set[frozenset[OntologyStringResource]], set[OntologyStringResource]]:
        all_conflicts = set()
        case_conflict_subsets = set()
        clean_resources = set()
        for potential_conflict_set in maybe_good_resources_by_active_syn_lower.values():

            if OntologyStringConflictAnalyser._resource_set_has_case_conflicts(
                potential_conflict_set, strict
            ):
                # uh ho - we have multiple identical synonyms attached to different resources
                case_conflict_subsets.add(frozenset(potential_conflict_set))
                all_conflicts.update(potential_conflict_set)
            else:
                clean_resources.update(potential_conflict_set)
        # there are scenarios wherein a case conflict can occur transitively.
        # Therefore, we need to eliminate from the clean resources any conflicted ones
        clean_resources.difference_update(all_conflicts)
        return case_conflict_subsets, clean_resources

    @staticmethod
    def build_synonym_defaultdict(
        resources: Iterable[OntologyStringResource],
    ) -> defaultdict[str, set[OntologyStringResource]]:
        maybe_good_resources_by_active_syn_lower = defaultdict(set)
        for resource in resources:
            for syn in resource.all_synonyms():
                maybe_good_resources_by_active_syn_lower[syn.text.lower()].add(resource)
        return maybe_good_resources_by_active_syn_lower

    def check_for_normalised_behaviour_conflicts_and_merge_if_possible(
        self, resources: set[OntologyStringResource]
    ) -> tuple[
        set[OntologyStringResource],
        set[OntologyStringResource],
        set[frozenset[OntologyStringResource]],
    ]:
        """Find behaviour conflicts in the resource set indexed by syn_norm.

        If possible, resources will be merged. If not, they will be added to a set of
        conflicting resources.

        :param resources:
        :return: A set of newly created merged resources, a set of resources eliminated
            through the merge, and a set of resources subsets that conflict and cannot
            be merged,
        """

        resources_by_syn_norm = (
            self._group_resources_by_syn_norm_and_check_for_normalisation_consistency_errors(
                resources
            )
        )

        normalisation_conflicts = set()
        merged_resources = set()
        eliminated_resources = set()
        for syn_norm, potentially_conflicting_resources in resources_by_syn_norm.items():
            if len(potentially_conflicting_resources) == 1:
                # no conflict
                continue

            # uh ho we have a normalisation/behaviour/id set conflict. If we can't merge, report an error
            original_synonyms_merged: set[Synonym] = set()
            generated_synonyms_merged: set[Synonym] = set()
            associated_id_sets_this_syn_norm: set[frozenset[EquivalentIdSet]] = set()
            comments = []
            behaviours = set()
            for conflicted_resource in potentially_conflicting_resources:

                behaviours.add(conflicted_resource.behaviour)
                original_synonyms_merged.update(conflicted_resource.original_synonyms)
                generated_synonyms_merged.update(conflicted_resource.alternative_synonyms)
                if conflicted_resource.associated_id_sets is not None:
                    associated_id_sets_this_syn_norm.add(conflicted_resource.associated_id_sets)
                if conflicted_resource.comment is not None:
                    comments.append(conflicted_resource.comment)

            if len(behaviours) > 1 or len(associated_id_sets_this_syn_norm) > 1:
                # uh ho - behaviours/id set clash. Add to norm conflicts
                normalisation_conflicts.add(frozenset(potentially_conflicting_resources))
            else:
                # merge the resources
                merged_resource = OntologyStringResource(
                    behaviour=next(iter(behaviours)),
                    original_synonyms=frozenset(original_synonyms_merged),
                    alternative_synonyms=frozenset(generated_synonyms_merged),
                    associated_id_sets=next(iter(associated_id_sets_this_syn_norm))
                    if len(associated_id_sets_this_syn_norm) == 1
                    else None,
                    comment="\n".join(comments) if len(comments) > 0 else None,
                )
                merged_resources.add(merged_resource)

                logger.warning(
                    "duplicate resource set merged. synonym_norm: %s, conflicts:\n%s\n merged to: %s",
                    syn_norm,
                    potentially_conflicting_resources,
                    merged_resource,
                )
                eliminated_resources.update(potentially_conflicting_resources)
        return merged_resources, eliminated_resources, normalisation_conflicts

    def _group_resources_by_syn_norm_and_check_for_normalisation_consistency_errors(
        self, resources: set[OntologyStringResource]
    ) -> defaultdict[str, set[OntologyStringResource]]:
        resources_by_syn_norm: defaultdict[str, set[OntologyStringResource]] = defaultdict(set)
        normalisation_errors = set()
        for resource in resources:
            syn_norms_this_resource = set(
                StringNormalizer.normalize(original_syn.text, entity_class=self.entity_class)
                for original_syn in resource.original_synonyms
            )
            if len(syn_norms_this_resource) > 1:
                normalisation_errors.add(resource)
            syn_norm = next(iter(syn_norms_this_resource))
            resources_by_syn_norm[syn_norm].add(resource)
        if normalisation_errors:
            # This should never be thrown by automatically generated resources
            raise CurationError(
                "One or more OntologyStringResource contains original synonyms that no longer normalise to the same value. This"
                " can happen if the implementation of the StringNormalizer has changed. Please correct the following"
                " resources:\n" + "\n".join(resource.to_json() for resource in normalisation_errors)
            )
        return resources_by_syn_norm

    @staticmethod
    def _resource_set_has_case_conflicts(
        resources: set[OntologyStringResource], strict: bool = False
    ) -> bool:
        """Checks for case conflicts in a set of :class:`.OntologyStringResource`\\ s.

        Intended behaviour is to allow different cases to produce different
        confidences based on confidence rank. A case-sensitive rank must always
        be higher than a case-insensitive rank, or a conflict will occur.

        For example, the following situation should be supported:

        "eGFR" -> ci and IGNORE
        "EGFR" -> cs and POSSIBLE
        "Egfr" -> cs and PROBABLE

        ...while the following situations are conflicted:

        "eGFR" -> ci and PROBABLE
        "Egfr" -> cs and POSSIBLE

        "Egfr" -> cs and PROBABLE
        "Egfr" -> cs and POSSIBLE

        "EGFR" -> ci and PROBABLE
        "Egfr" -> ci and POSSIBLE

        "Egfr" -> ci and PROBABLE
        "Egfr" -> ci and POSSIBLE

        :param resources:
        :param strict: if True, then the function will return True if there
            are multiple mention confidences for a given string, regardless of case sensitivity
        :return:
        """
        cs_conf_lookup = defaultdict(set)
        ci_conf_lookup = defaultdict(set)
        for resource in resources:
            for syn in resource.active_ner_synonyms():
                if syn.case_sensitive:
                    cs_conf_lookup[syn.text].add(syn.mention_confidence)
                else:
                    ci_conf_lookup[syn.text.lower()].add(syn.mention_confidence)

        for cased_syn_string, cs_confidences in cs_conf_lookup.items():
            ci_confidences: set[MentionConfidence] = ci_conf_lookup.get(
                cased_syn_string.lower(), set()
            )
            if len(ci_confidences) > 1 or (
                len(ci_confidences) == 1
                and len(cs_confidences) > 0
                and (min(cs_confidences) <= min(ci_confidences) or strict)
            ):
                return True
        for ci_confidences in ci_conf_lookup.values():
            if len(ci_confidences) > 1:
                return True
        return False

    def merge_human_and_auto_resources(
        self,
        human_curated_resources: set[OntologyStringResource],
        autocurated_resources: set[OntologyStringResource],
    ) -> OntologyResourceSetMergeReport:
        """Merge a set of human curated resources with a set of automatically curated
        resources, preferring the human set where possible.

        Note that the output is not guaranteed to be conflict free - consider calling
        :meth:`~.verify_resource_set_integrity`.

        :param human_curated_resources:
        :param autocurated_resources:
        :return:
        """
        # normalisation errors should have already been checked with self.verify_resource_set_integrity before calling this

        obsolete_human_set = set()
        resources_with_discrepancies = set()
        superfluous_human_curations = human_curated_resources.intersection(autocurated_resources)
        human_by_syn_norm = {
            resource.syn_norm_for_linking(self.entity_class): resource
            for resource in human_curated_resources
        }
        default_by_syn_norm = {
            resource.syn_norm_for_linking(self.entity_class): resource
            for resource in autocurated_resources
        }

        working_dict = {}
        for default_syn_norm, default_resource in default_by_syn_norm.items():

            maybe_human_curation = human_by_syn_norm.pop(default_syn_norm, None)
            if maybe_human_curation is None:
                working_dict[default_syn_norm] = default_resource
            else:
                working_dict[default_syn_norm] = maybe_human_curation
                # check the set of strings are equivalent. Otherwise new ones may have been created by syn generation
                if set(syn.text for syn in maybe_human_curation.original_synonyms) != set(
                    syn.text for syn in default_resource.original_synonyms
                ) or set(syn.text for syn in maybe_human_curation.alternative_synonyms) != set(
                    syn.text for syn in default_resource.alternative_synonyms
                ):
                    resources_with_discrepancies.add(
                        (
                            maybe_human_curation,
                            default_resource,
                        )
                    )

        # add any remaining to the obsolete set, unless they're additional
        for human_syn_norm, human_curation in human_by_syn_norm.items():
            if human_curation.additional_to_source:
                working_dict[human_syn_norm] = human_curation
            else:
                obsolete_human_set.add(human_curation)

        return OntologyResourceSetMergeReport(
            effective_resources=set(working_dict.values()),
            superfluous_resources=superfluous_human_curations,
            resources_with_discrepancies=resources_with_discrepancies,
            obsolete_resources=obsolete_human_set,
        )


class LinkingCandidateModificationResult(AutoNameEnum):
    ID_SET_MODIFIED = auto()
    LINKING_CANDIDATE_ADDED = auto()
    LINKING_CANDIDATE_DROPPED = auto()
    NO_ACTION = auto()


class OntologyResourceProcessor:
    """A OntologyResourceProcessor is responsible for modifying the set of
    :class:`.LinkingCandidate`\\s produced by an
    :class:`kazu.ontology_preprocessing.base.OntologyParser` with any relevant
    :class:`.GlobalParserActions` and/or :class:`.OntologyStringResource` associated
    with the parser.

    This class should be used before instances of :class:`.LinkingCandidate`\\s are loaded into the
    internal database representation.
    """

    # OntologyStringBehaviours are applied in the following order
    BEHAVIOUR_APPLICATION_ORDER = (
        OntologyStringBehaviour.ADD_FOR_NER_AND_LINKING,
        OntologyStringBehaviour.ADD_FOR_LINKING_ONLY,
        OntologyStringBehaviour.DROP_FOR_LINKING,
    )
    _BEHAVIOUR_TO_ORDER_INDEX = {behav: i for i, behav in enumerate(BEHAVIOUR_APPLICATION_ORDER)}

    def __init__(
        self,
        parser_name: str,
        entity_class: str,
        global_actions: Optional[GlobalParserActions],
        resources: list[OntologyStringResource],
        linking_candidates: set[LinkingCandidate],
    ):
        """

        :param parser_name: name of parser to process
        :param entity_class: name of parser entity_class to process (typically as passed to :class:`kazu.ontology_preprocessing.base.OntologyParser`\\ )
        :param global_actions:
        :param resources:
        :param linking_candidates:
        """
        self.global_actions = global_actions
        self.entity_class = entity_class
        self.parser_name = parser_name
        self._candidates_by_syn_norm: dict[NormalisedSynonymStr, LinkingCandidate] = {}
        self._candidates_by_id: defaultdict[Idx, set[LinkingCandidate]] = defaultdict(set)
        for candidate in linking_candidates:
            self._update_candidate_lookups(candidate, False)
        self.resources = set(resources)
        self.dropped_keys: set[NormalisedSynonymStr] = set()

    @classmethod
    def resource_sort_key(cls, resource: OntologyStringResource) -> tuple[int, bool]:
        """Determines the order resources are processed in.

        We use associated_id_sets as a key, so that any overrides will be processed
        after any original behaviours.
        """
        return (
            cls._BEHAVIOUR_TO_ORDER_INDEX[resource.behaviour],
            resource.associated_id_sets is not None,
        )

    def _update_candidate_lookups(
        self, candidate: LinkingCandidate, override: bool
    ) -> Literal[
        LinkingCandidateModificationResult.LINKING_CANDIDATE_ADDED,
        LinkingCandidateModificationResult.NO_ACTION,
    ]:

        safe_to_add = False
        maybe_existing_candidate = self._candidates_by_syn_norm.get(candidate.synonym_norm)
        if maybe_existing_candidate is None:
            logger.debug("adding new candidate %s", candidate)
            safe_to_add = True
        elif override:
            safe_to_add = True
            logger.debug("overriding existing candidate %s", maybe_existing_candidate)
        elif (
            len(
                candidate.associated_id_sets.symmetric_difference(
                    maybe_existing_candidate.associated_id_sets
                )
            )
            > 0
        ):
            logger.warning(
                "conflict on synonym norms \n%s\n%s\nthe latter will be ignored",
                maybe_existing_candidate,
                candidate,
            )
        if safe_to_add:
            self._candidates_by_syn_norm[candidate.synonym_norm] = candidate
            for equiv_ids in candidate.associated_id_sets:
                for idx in equiv_ids.ids:
                    self._candidates_by_id[idx].add(candidate)
            return LinkingCandidateModificationResult.LINKING_CANDIDATE_ADDED
        else:
            return LinkingCandidateModificationResult.NO_ACTION

    def _drop_linking_candidate(self, synonym: NormalisedSynonymStr) -> None:
        """Remove a linking candidate from the database, so that it cannot be used as a
        linking target.

        :param synonym:
        :return:
        """
        try:
            candidate_to_remove = self._candidates_by_syn_norm.pop(synonym)
            self.dropped_keys.add(synonym)
            for equiv_id_set in candidate_to_remove.associated_id_sets:
                for idx in equiv_id_set.ids:
                    candidates_by_id = self._candidates_by_id.get(idx)
                    if candidates_by_id is not None:
                        candidates_by_id.remove(candidate_to_remove)
            logger.debug(
                "successfully dropped %s from database for %s",
                synonym,
                self.entity_class,
            )
        except KeyError:
            if synonym in self.dropped_keys:
                logger.debug(
                    "tried to drop %s from database, but key already dropped by another OntologyStringResource for %s",
                    synonym,
                    self.parser_name,
                )
            else:
                logger.warning(
                    "tried to drop %s from database, but key doesn't exist for %s",
                    synonym,
                    self.parser_name,
                )

    def _drop_id_from_all_linking_candidates(
        self, id_to_drop: Idx
    ) -> Counter[
        Literal[
            LinkingCandidateModificationResult.ID_SET_MODIFIED,
            LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED,
            LinkingCandidateModificationResult.NO_ACTION,
        ]
    ]:
        """Remove a given id from all :class:`.LinkingCandidate`\\ s.

        Drop any :class:`.LinkingCandidate`\\ s with no remaining ID after removal.

        :param id_to_drop:
        :return:
        """

        candidates_to_modify = self._candidates_by_id.get(id_to_drop, set())
        counter = Counter(
            self._drop_id_from_linking_candidate(
                id_to_drop=id_to_drop, candidate_to_modify=candidate_to_modify
            )
            for candidate_to_modify in set(candidates_to_modify)
        )

        return counter

    def _drop_id_from_linking_candidate(
        self, id_to_drop: Idx, candidate_to_modify: LinkingCandidate
    ) -> Literal[
        LinkingCandidateModificationResult.ID_SET_MODIFIED,
        LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED,
        LinkingCandidateModificationResult.NO_ACTION,
    ]:
        """Remove an id from a given :class:`.LinkingCandidate`\\ .

        :param id_to_drop:
        :param candidate_to_modify:
        :return:
        """
        new_assoc_id_frozenset = self._drop_id_from_associated_id_sets(
            id_to_drop, candidate_to_modify.associated_id_sets
        )
        if (
            len(new_assoc_id_frozenset.symmetric_difference(candidate_to_modify.associated_id_sets))
            == 0
        ):
            return LinkingCandidateModificationResult.NO_ACTION
        else:
            return self._modify_or_drop_linking_candidate_after_id_set_change(
                new_associated_id_sets=new_assoc_id_frozenset, candidate=candidate_to_modify
            )

    def _drop_id_from_associated_id_sets(
        self, id_to_drop: Idx, associated_id_sets: AssociatedIdSets
    ) -> AssociatedIdSets:
        """Remove an id from a :class:`.AssociatedIdSets`\\ .

        :param id_to_drop:
        :param associated_id_sets:
        :return:
        """
        new_assoc_id_set = set()
        for equiv_id_set in associated_id_sets:
            if id_to_drop in equiv_id_set.ids:
                updated_ids_and_source = frozenset(
                    id_tup for id_tup in equiv_id_set.ids_and_source if id_tup[0] != id_to_drop
                )
                if len(updated_ids_and_source) > 0:
                    updated_equiv_id_set = EquivalentIdSet(updated_ids_and_source)
                    new_assoc_id_set.add(updated_equiv_id_set)
            else:
                new_assoc_id_set.add(equiv_id_set)
        new_assoc_id_frozenset = frozenset(new_assoc_id_set)
        return new_assoc_id_frozenset

    def _modify_or_drop_linking_candidate_after_id_set_change(
        self, new_associated_id_sets: AssociatedIdSets, candidate: LinkingCandidate
    ) -> Literal[
        LinkingCandidateModificationResult.ID_SET_MODIFIED,
        LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED,
    ]:
        """Modifies or drops a :class:`.LinkingCandidate` after a
        :class:`.AssociatedIdSets` has changed.

        :param new_associated_id_sets:
        :param candidate:
        :return:
        """
        result: Literal[
            LinkingCandidateModificationResult.ID_SET_MODIFIED,
            LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED,
        ]
        if len(new_associated_id_sets) > 0:
            if new_associated_id_sets == candidate.associated_id_sets:
                raise ValueError(
                    "function called inappropriately where the id sets haven't changed. This"
                    " has failed as it will otherwise modify the value of aggregated_by, when"
                    " nothing has changed"
                )
            new_candidate = dataclasses.replace(
                candidate,
                associated_id_sets=new_associated_id_sets,
                aggregated_by=EquivalentIdAggregationStrategy.MODIFIED_BY_CURATION,
            )
            add_result = self._update_candidate_lookups(new_candidate, True)
            assert add_result is LinkingCandidateModificationResult.LINKING_CANDIDATE_ADDED
            result = LinkingCandidateModificationResult.ID_SET_MODIFIED
        else:
            # if there are no longer any id sets associated with the record, remove it completely
            self._drop_linking_candidate(candidate.synonym_norm)
            result = LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED
        return result

    def export_resources_and_final_candidates(
        self,
    ) -> tuple[list[OntologyStringResource], set[LinkingCandidate]]:
        """Perform any updates required to the linking candidates as specified in the
        curations/global actions.

        The returned :class:`.OntologyStringResource`\\s can be used for Dictionary based NER, whereas the
        returned :class:`.LinkingCandidate`\\s can be loaded into the internal database for linking.

        :return:
        """
        self._process_global_actions()
        return list(self._process_resources()), set(self._candidates_by_syn_norm.values())

    def _process_resources(self) -> Iterable[OntologyStringResource]:
        for resource in sorted(self.resources, key=self.resource_sort_key):
            maybe_resource = self._process_resource_action(resource)
            if maybe_resource:
                yield maybe_resource

    def _process_resource_action(
        self, resource: OntologyStringResource
    ) -> Optional[OntologyStringResource]:

        if resource.behaviour is OntologyStringBehaviour.DROP_FOR_LINKING:
            self._drop_linking_candidate(resource.syn_norm_for_linking(self.entity_class))
            result = None
        elif resource.behaviour is OntologyStringBehaviour.ADD_FOR_LINKING_ONLY:
            result = self._attempt_to_add_database_entry_for_resource(
                resource,
            )
        elif resource.behaviour is OntologyStringBehaviour.ADD_FOR_NER_AND_LINKING:
            result = self._attempt_to_add_database_entry_for_resource(resource)
        else:
            raise ValueError(f"unknown behaviour for parser {self.parser_name}, {resource}")
        if result is LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED:
            return None
        else:
            return resource

    def _process_global_actions(self) -> None:
        if self.global_actions is None:
            return None

        override_resources_by_id = defaultdict(set)
        for resource in self.resources:
            if resource.associated_id_sets is not None:
                for equiv_id_set in resource.associated_id_sets:
                    for idx in equiv_id_set.ids:
                        override_resources_by_id[idx].add(resource)

        for action in self.global_actions.parser_behaviour(self.parser_name):
            if action.behaviour is ParserBehaviour.DROP_IDS_FROM_PARSER:
                ids = action.parser_to_target_id_mappings[self.parser_name]
                for idx in ids:
                    counter_this_idx = self._drop_id_from_all_linking_candidates(idx)
                    if (
                        counter_this_idx[LinkingCandidateModificationResult.ID_SET_MODIFIED]
                        + counter_this_idx[
                            LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED
                        ]
                        == 0
                    ):
                        logger.warning("failed to drop %s from %s", idx, self.parser_name)
                    else:
                        logger.debug(
                            "dropped ID %s from %s. LinkingCandidate modified count: %s, LinkingCandidate dropped count: %s",
                            idx,
                            self.parser_name,
                            counter_this_idx[LinkingCandidateModificationResult.ID_SET_MODIFIED],
                            counter_this_idx[
                                LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED
                            ],
                        )

                        for override_resource_to_modify in set(
                            override_resources_by_id.get(idx, set())
                        ):
                            assert override_resource_to_modify.associated_id_sets is not None
                            new_associated_id_sets = self._drop_id_from_associated_id_sets(
                                idx, override_resource_to_modify.associated_id_sets
                            )
                            if len(new_associated_id_sets) == 0:

                                self.resources.remove(override_resource_to_modify)
                                override_resources_by_id[idx].remove(override_resource_to_modify)
                                logger.debug(
                                    "removed resource %s because of global action",
                                    override_resource_to_modify,
                                )
                            elif (
                                new_associated_id_sets
                                != override_resource_to_modify.associated_id_sets
                            ):
                                self.resources.remove(override_resource_to_modify)
                                override_resources_by_id[idx].remove(override_resource_to_modify)
                                mod_resource = dataclasses.replace(
                                    override_resource_to_modify,
                                    associated_id_sets=new_associated_id_sets,
                                )
                                self.resources.add(mod_resource)
                                override_resources_by_id[idx].add(mod_resource)
                                logger.debug(
                                    "modified resource %s to %s because of global action",
                                    override_resource_to_modify,
                                    mod_resource,
                                )

            else:
                raise ValueError(f"unknown behaviour for parser {self.parser_name}, {action}")
        return None

    def _attempt_to_add_database_entry_for_resource(
        self,
        resource: OntologyStringResource,
    ) -> Literal[
        LinkingCandidateModificationResult.LINKING_CANDIDATE_ADDED,
        LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED,
        LinkingCandidateModificationResult.NO_ACTION,
    ]:
        """Create a new :class:`~kazu.data.LinkingCandidate` for the database, or return
        an existing matching one if already present.

        Notes:

        If a syn_norm already exists in self._candidates_by_syn_norm that matches 'resource.syn_norm_for_linking',
        this method will check to see if the 'resource_associated_id_set' matches the existing terms
        :class:`.AssociatedIdSets`\\.

        If so, no action will be taken. If not, a warning will be logged as adding it
        will cause irregularities in the database.

        If the syn_norm does not exist, this method will create a new :class:`~kazu.data.LinkingCandidate`
        with the provided :class:`.AssociatedIdSets`\\.

        :param resource:
        :return:
        """
        log_prefix = "%(parser_name)s attempting to create linking candidate for syn_norm: <%(syn_norm)s> resource: %(resource)s"
        syn_norm = resource.syn_norm_for_linking(self.entity_class)
        log_formatting_dict: dict[str, Any] = {
            "parser_name": self.parser_name,
            "syn_norm": syn_norm,
            "resource": resource,
        }

        # look up the syn norm in the db
        maybe_existing_linking_candidate = self._candidates_by_syn_norm.get(syn_norm)
        resource_associated_id_set = resource.associated_id_sets
        if resource_associated_id_set is None and maybe_existing_linking_candidate is not None:
            logger.debug(
                log_prefix
                + " but no associated id set provided, so candidate will inherit the parser defaults",
                log_formatting_dict,
            )
            return LinkingCandidateModificationResult.NO_ACTION
        elif resource_associated_id_set is None and maybe_existing_linking_candidate is None:
            logger.info(
                log_prefix
                + " but syn_norm <%(syn_norm)s> does not exist in synonym database."
                + " It may have been removed by a global action."
                + " Since no id set was provided, no entry can be created",
                log_formatting_dict,
            )
            return LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED

        # resource_associated_id_set is implicitly not None
        assert resource_associated_id_set is not None
        if len(resource_associated_id_set) == 0:
            logger.debug(
                "all ids removed by global action for %s,  Parser name: %s",
                resource,
                self.parser_name,
            )
            return LinkingCandidateModificationResult.LINKING_CANDIDATE_DROPPED

        if maybe_existing_linking_candidate is not None:
            log_formatting_dict[
                "existing_id_set"
            ] = maybe_existing_linking_candidate.associated_id_sets

            if (
                len(
                    resource_associated_id_set.symmetric_difference(
                        maybe_existing_linking_candidate.associated_id_sets
                    )
                )
                == 0
            ):
                logger.debug(
                    log_prefix
                    + " but syn_nom <%(syn_norm)s> already exists in synonym database."
                    + "since this LinkingCandidate matches the id_set, no action is required. %(existing_id_set)s",
                    log_formatting_dict,
                )
                return LinkingCandidateModificationResult.NO_ACTION
            else:
                logger.debug(
                    log_prefix
                    + " . Will remove existing syn_norm <%(syn_norm)s> as an ID set override has been specified",
                    log_formatting_dict,
                )

        # no candidate exists, or we want to override so one will be made
        assert resource_associated_id_set is not None
        for equiv_id_set in set(resource_associated_id_set):
            for idx in equiv_id_set.ids:
                if idx not in self._candidates_by_id:
                    resource_associated_id_set = self._drop_id_from_associated_id_sets(
                        id_to_drop=idx, associated_id_sets=resource_associated_id_set
                    )
                    logger.warning(
                        "Attempted to add candidate containing %s but this id does not exist in the parser and will be ignored",
                        idx,
                    )
        if len(resource_associated_id_set) > 0:
            is_symbolic = any(
                StringNormalizer.classify_symbolic(syn.text, self.entity_class)
                for syn in resource.original_synonyms
            )
            new_candidate = LinkingCandidate(
                synonym_norm=syn_norm,
                raw_synonyms=frozenset(syn.text for syn in resource.original_synonyms),
                is_symbolic=is_symbolic,
                mapping_types=frozenset(("kazu_curated",)),
                associated_id_sets=resource_associated_id_set,
                parser_name=self.parser_name,
                aggregated_by=EquivalentIdAggregationStrategy.MODIFIED_BY_CURATION,
            )
            return self._update_candidate_lookups(new_candidate, True)
        else:
            return LinkingCandidateModificationResult.NO_ACTION



import logging
from typing import Iterable

import streamlit as st
from kazu.krt.components import (
    ResourceEditor,
    ParserSelector,
    PlaceholderResource,
)
from kazu.krt.string_editor.utils import (
    CaseConflictResolutionRequest,
    ResourceConflict,
    ResourceConflictManager,
)
from kazu.krt.utils import get_resource_manager


@st.cache_resource
def get_manager() -> ResourceConflictManager:
    return ResourceConflictManager(manager=get_resource_manager())


def reset() -> None:
    # type ignore as mypy doesn't pick up annotations
    get_resource_manager.clear()  # type: ignore[attr-defined]
    get_manager.clear()  # type: ignore[attr-defined]


class StringConflictForm:
    STRING_CONFLICT_BATCH_APPLY = "STRING_CONFLICT_BATCH_APPLY"
    DATAFRAME_SELECTION = "DATAFRAME_SELECTION"
    DATAFRAME = "DATAFRAME"
    FORM_PICKER = "FORM_PICKER"

    @staticmethod
    def _submit_form_for_batch_conflict_resolution(conflicts: Iterable[ResourceConflict]) -> None:
        for conflict in conflicts:
            logging.info(f"submit form {id(conflict)}")
            if (
                st.session_state.get(StringConflictForm.STRING_CONFLICT_BATCH_APPLY)
                is CaseConflictResolutionRequest.OPTIMISTIC
            ):
                conflict.batch_resolve(True)
            elif (
                st.session_state.get(StringConflictForm.STRING_CONFLICT_BATCH_APPLY)
                is CaseConflictResolutionRequest.PESSIMISTIC
            ):
                conflict.batch_resolve(False)
            flow = get_manager()
            flow.sync_resources_for_resolved_resource_conflict_and_find_new_conflicts(conflict)

        if "submit_batch" in st.session_state:
            del st.session_state["submit_batch"]
        del st.session_state[StringConflictForm.DATAFRAME]

    @staticmethod
    def _submit_form_for_individual_conflicts(conflicts: Iterable[ResourceConflict]) -> None:
        for conflict in conflicts:
            logging.info(f"submit form {id(conflict)}")
            if (
                st.session_state.get(StringConflictForm.STRING_CONFLICT_BATCH_APPLY)
                is CaseConflictResolutionRequest.OPTIMISTIC
            ):
                conflict.batch_resolve(True)
            elif (
                st.session_state.get(StringConflictForm.STRING_CONFLICT_BATCH_APPLY)
                is CaseConflictResolutionRequest.PESSIMISTIC
            ):
                conflict.batch_resolve(False)
            else:
                for parser_name, resource_dict in conflict.parser_to_resource_to_resolution.items():

                    for (
                        original_resource,
                        new_resource,
                    ) in ResourceEditor.extract_form_data_from_state(
                        parser_name=parser_name, resources=resource_dict.keys()
                    ):
                        conflict.parser_to_resource_to_resolution[parser_name][
                            original_resource
                        ] = new_resource
            flow = get_manager()
            flow.sync_resources_for_resolved_resource_conflict_and_find_new_conflicts(conflict)

        if "submit_batch" in st.session_state:
            del st.session_state["submit_batch"]
        del st.session_state[StringConflictForm.DATAFRAME]

    @staticmethod
    def _individual_conflict_resolution_form() -> None:
        row_ids = (
            st.session_state[StringConflictForm.DATAFRAME_SELECTION]
            .get("selection", {})
            .get("rows", None)
        )
        if len(row_ids) == 1:
            row_id = row_ids[0]
            df = st.session_state[StringConflictForm.DATAFRAME]
            conflict_id = df.iloc[[row_id]]["id"].values[0]
            conflict = get_manager().unresolved_conflicts[conflict_id]
            resources = set(
                resource
                for resource_dict in conflict.parser_to_resource_to_resolution.values()
                for resource in resource_dict
            )
            ResourceEditor.display_resource_editor(
                resources=resources,
                on_click_override=StringConflictForm._submit_form_for_individual_conflicts,
                args=([conflict],),
            )

    @staticmethod
    def _batch_conflict_resolution_form() -> None:
        df = st.session_state[StringConflictForm.DATAFRAME]
        resolution_choices = list(CaseConflictResolutionRequest)
        resolution_choices.remove(CaseConflictResolutionRequest.CUSTOM)
        st.radio(
            "select a resolution",
            options=resolution_choices,
            key=StringConflictForm.STRING_CONFLICT_BATCH_APPLY,
        )
        st.write("select rows to apply the resolution to")
        event = st.dataframe(
            df,
            use_container_width=True,
            selection_mode="multi-row",
            on_select="rerun",
            hide_index=True,
            column_config={"id": None},
            key=StringConflictForm.DATAFRAME_SELECTION,
        )
        conflicts = []
        for row_id in event.get("selection", {}).get("rows", []):
            conflict_id = df.iloc[[row_id]]["id"].values[0]
            conflict = get_manager().unresolved_conflicts[conflict_id]
            conflicts.append(conflict)

        if conflicts:
            disabled = False
        else:
            disabled = True
        submitted = st.button("submit batch", key="submit_batch", disabled=disabled)
        if submitted:
            StringConflictForm._submit_form_for_batch_conflict_resolution(conflicts)
            st.rerun()

    @staticmethod
    def resolve_conflicts_form() -> None:
        with st.container(border=True):
            st.write(f"""{len(get_manager().unresolved_conflicts)} remaining discrepancies""")
            st.radio(
                "mode",
                options=["apply to all", "edit individual"],
                index=0,
                key=StringConflictForm.FORM_PICKER,
            )
            if StringConflictForm.DATAFRAME not in st.session_state:
                maybe_df = get_manager().summary_df()
                if not maybe_df.empty:
                    maybe_df = maybe_df.sort_values(by="string len", ascending=False)
                    st.session_state[StringConflictForm.DATAFRAME] = maybe_df
                    st.rerun()

            elif st.session_state[StringConflictForm.DATAFRAME].empty:
                st.write("no conflicts found")
            else:

                if st.session_state[StringConflictForm.FORM_PICKER] == "apply to all":
                    StringConflictForm._batch_conflict_resolution_form()
                else:
                    df = st.session_state[StringConflictForm.DATAFRAME]
                    st.dataframe(
                        df,
                        use_container_width=True,
                        selection_mode="single-row",
                        on_select="rerun",
                        hide_index=True,
                        column_config={"conflict": None},
                        key=StringConflictForm.DATAFRAME_SELECTION,
                    )
                    StringConflictForm._individual_conflict_resolution_form()

    @staticmethod
    def search_and_edit_synonyms_form() -> None:
        with st.container(border=True):
            text = st.text_input("search for a string")
            if text:
                manager = get_resource_manager()
                maybe_resources = manager.synonym_lookup.get(text.lower())
                if not maybe_resources:
                    st.write("No existing resources found that contain this string")
                    StringConflictForm._add_new_resource_form(text)
                else:
                    st.write(
                        "One or more resources already exists for this string. You can edit them here:"
                    )
                    mode = st.radio(
                        label="edit existing or create new",
                        options=["edit existing", "create new"],
                        index=0,
                    )
                    if mode == "edit existing":
                        ResourceEditor.display_resource_editor(maybe_resources)
                    else:
                        StringConflictForm._add_new_resource_form(text)

    @staticmethod
    def _add_new_resource_form(text: str) -> None:
        ParserSelector.display_parser_selector(clear_state=False)
        maybe_parser_name = ParserSelector.get_selected_parser_name()
        if maybe_parser_name:
            parser = get_resource_manager().parsers[maybe_parser_name]
            st.write(f"selected parser is {parser.name}, entity class: {parser.entity_class}")
            PlaceholderResource.create_placeholder_resource(text=text, parser_name=parser.name)
            parser_name, placeholder_resource = PlaceholderResource.get_placeholder_resource()
            ResourceEditor.display_resource_editor(
                resources={placeholder_resource},
                maybe_parser_name=parser_name,
                on_click_override=ResourceEditor.submit_form_for_addition,
            )



from typing import Optional, Iterable

import pandas as pd
from kazu.data import OntologyStringResource
from kazu.krt.resource_manager import ResourceManager
from kazu.krt.utils import (
    create_new_resource_with_updated_synonyms,
    resource_to_df,
)


class SynonymDiscrepancy:
    """This class represents a discrepancy between a human-generated
    :class:`.OntologyStringResource` and an auto-generated
    :class:`.OntologyStringResource`.

    It provides methods to automatically resolve the discrepancy, convert the resources
    to a DataFrame, and get an example string for display.
    """

    def __init__(
        self, human_resource: OntologyStringResource, auto_resource: OntologyStringResource
    ):
        self.auto_resource = auto_resource
        self.human_resource = human_resource

    def auto_resolve(self) -> Optional[OntologyStringResource]:
        """This method attempts to automatically resolve discrepancies between human and
        auto resources.

        It first creates a set of tuples containing the mention confidence and case
        sensitivity for all :class:`.Synonym`\\s in the human resource. If there is more than one
        unique tuple in the set, it means there are discrepancies in the human resource
        itself, and the method returns None. If there is exactly one unique tuple, it
        means all synonyms in the human resource have the same mention confidence and
        case sensitivity. In this case, it updates all forms of the auto resource with
        this mention confidence and case sensitivity, and returns the updated auto
        resource.

        :return: The updated auto resource if discrepancies can be auto-resolved, which
            can be used as a human override. None otherwise.
        """
        human_aspects = set(
            (
                x.mention_confidence,
                x.case_sensitive,
            )
            for x in self.human_resource.all_synonyms()
        )
        if len(human_aspects) != 1:
            return None
        else:
            new_conf, new_cs = next(iter(human_aspects))
            return create_new_resource_with_updated_synonyms(
                new_conf=new_conf, new_cs=new_cs, resource=self.auto_resource
            )

    def dataframe(self) -> pd.DataFrame:
        """Converts the human and auto resources to DataFrames, merges them, and returns
        the rows with any null values (i.e. discrepancies)

        :return: A :class:`~pandas.DataFrame` representing the discrepancies between the human and auto
            resources.
        """
        human_df = resource_to_df(self.human_resource)
        auto_df = resource_to_df(self.auto_resource)
        merged = pd.merge(
            human_df, auto_df, how="outer", on=["type", "text"], suffixes=("_human", "_auto")
        )
        return merged[merged.isnull().any(axis=1)]

    def example_string(self) -> str:
        """Returns an example string from the human resource's original synonyms."""
        return next(iter(self.human_resource.original_synonyms)).text


class ResourceDiscrepancyManger:
    """This class manages :class:`.SynonymDiscrepancy`\\s between human-generated
    resources and auto- generated resources.

    It provides methods to automatically resolve all discrepancies, commit changes to
    the resources, and get a summary DataFrame.
    """

    def __init__(
        self,
        parser_name: str,
        manager: ResourceManager,
    ):
        """Initializes the ResourceDiscrepancyManager.

        :param parser_name: The name of the parser used to generate the resources.
        :param manager: The :class:`.ResourceManager` object used to manage the resources.
        """
        self.manager = manager
        self.parser_name = parser_name
        report = manager.parser_to_report[parser_name]
        if not report.merge_report:
            todo = set()
        else:

            todo = set(
                SynonymDiscrepancy(human_resource=human_curation, auto_resource=autocuration)
                for human_curation, autocuration in report.merge_report.resources_with_discrepancies
            )
        self._build_discrepancy_lookup(todo)

    def _build_discrepancy_lookup(self, todo: Iterable[SynonymDiscrepancy]) -> None:
        self.unresolved_discrepancies: dict[int, SynonymDiscrepancy] = {
            i: discrepancy for i, discrepancy in enumerate(todo)
        }

    def apply_autofix_to_all(self) -> None:
        """Attempts to automatically resolve all discrepancies.

        If a discrepancy can be auto-resolved, it syncs the resources and removes the
        discrepancy from the internal unresolved discrepancies list.
        """
        for i in list(self.unresolved_discrepancies):
            discrepancy = self.unresolved_discrepancies[i]
            maybe_new_resource = discrepancy.auto_resolve()
            if maybe_new_resource is not None:
                self.manager.sync_resources(
                    original_resource=discrepancy.human_resource,
                    new_resource=maybe_new_resource,
                    parser_name=self.parser_name,
                )
                self.unresolved_discrepancies.pop(i)
        # need to recalculate lookup as UI indices no longer valid
        self._build_discrepancy_lookup(self.unresolved_discrepancies.values())

    def commit(
        self,
        original_human_resource: OntologyStringResource,
        new_resource: OntologyStringResource,
        index: int,
    ) -> None:
        """Commits changes to the resources and removes the discrepancy from the
        internal todo list.

        :param original_human_resource: The original human-generated resource.
        :param new_resource: The new resource to replace the original one.
        :param index: The index of the discrepancy.
        """
        self.manager.sync_resources(
            original_resource=original_human_resource,
            new_resource=new_resource,
            parser_name=self.parser_name,
        )
        self.unresolved_discrepancies.pop(index)
        self._build_discrepancy_lookup(self.unresolved_discrepancies.values())

    def summary_df(self) -> pd.DataFrame:
        """Returns a :class:`pandas.DataFrame` summarizing the unresolved discrepancies.

        :return: A DataFrame with columns for id, example text, and the number of unique
            synonyms in the human and auto resources.
        """
        data = []
        for i, discrepancy in self.unresolved_discrepancies.items():
            data.append(
                {
                    "id": i,
                    "example_text": discrepancy.example_string(),
                    "human_resource_unique_synonyms": len(
                        set(discrepancy.human_resource.all_synonyms())
                    ),
                    "auto_resource_unique_synonyms": len(
                        set(discrepancy.auto_resource.all_synonyms())
                    ),
                }
            )
        return pd.DataFrame(data)



import logging
from collections import defaultdict
from pathlib import Path
from typing import Optional, Iterable

from hydra.utils import instantiate
from kazu.data import (
    OntologyStringResource,
)
from kazu.ontology_preprocessing.base import OntologyParser
from kazu.ontology_preprocessing.curation_utils import (
    OntologyStringConflictAnalyser,
    dump_ontology_string_resources,
    OntologyResourceSetCompleteReport,
)
from kazu.pipeline import Pipeline
from kazu.krt import load_config


class ResourceManager:
    """The ResourceManager class is responsible for managing resources in the streamlit
    application.

    It manages the global state of all :class:`.OntologyStringResource`\\s.
    It is also responsible for saving and updates to these resources in the configured model pack.
    """

    def __init__(self, parsers: Iterable[OntologyParser]) -> None:
        """Initializes the ResourceManager instance.

        Sets up dictionaries for managing resources and parsers, and loads resources
        from each parser.
        """
        self.parser_to_curations: defaultdict[str, set[OntologyStringResource]] = defaultdict(set)
        # since duplicate resources may exist in multiple parsers, this mapping controls that
        self.resource_to_parsers: defaultdict[OntologyStringResource, set[str]] = defaultdict(set)
        # where to save the serialised resources to
        self.parser_to_path: dict[str, Path] = {}
        # parser to report
        self.parser_to_report: dict[str, OntologyResourceSetCompleteReport] = {}
        self.parsers: dict[str, OntologyParser] = {}
        self.parser_cache_invalid: set[str] = set()

        for parser in parsers:
            self.parsers[parser.name] = parser
            logging.info(f"loading data from parser: {parser.name}")
            if parser.curations_path is None:
                logging.warning(
                    "Parser %s has no curations path and will not be loaded", parser.name
                )
                continue
            self.parser_to_path[parser.name] = parser.curations_path
            _, resource_report = parser.populate_metadata_db_and_resolve_string_resources()
            self.parser_to_report[parser.name] = resource_report
            # we need the clean resources and the conflicted resources from the parser
            for resource in resource_report.final_conflict_report.clean_resources:
                self.resource_to_parsers[resource].add(parser.name)
            for resource_set in resource_report.final_conflict_report.case_conflicts:
                for resource in resource_set:
                    self.resource_to_parsers[resource].add(parser.name)

            if resource_report.human_conflict_report:
                for resource in resource_report.human_conflict_report.clean_resources:
                    self.resource_to_parsers[resource].add(parser.name)
                    self.parser_to_curations[parser.name].add(resource)
                for resource_set in resource_report.human_conflict_report.case_conflicts:
                    for resource in resource_set:
                        self.resource_to_parsers[resource].add(parser.name)
                        self.parser_to_curations[parser.name].add(resource)
                for (
                    frozen_resource_set
                ) in resource_report.human_conflict_report.normalisation_conflicts:
                    for resource in frozen_resource_set:
                        self.resource_to_parsers[resource].add(parser.name)
                        self.parser_to_curations[parser.name].add(resource)

            if resource_report.merge_report:
                for r1, r2 in resource_report.merge_report.resources_with_discrepancies:
                    self.resource_to_parsers[r1].add(parser.name)
                    self.resource_to_parsers[r2].add(parser.name)

        logging.info("building synonym lookup...")

        self.synonym_lookup = OntologyStringConflictAnalyser.build_synonym_defaultdict(
            self.resource_to_parsers.keys()
        )

    def parser_count(self) -> int:
        """Returns the number of parsers loaded by the ResourceManager instance.

        :return:
        """

        return len(self.parsers)

    def delete_resource(self, resource: OntologyStringResource, parser_name: str) -> None:
        """Deletes a resource from the internal state.

        :param resource: The resource to be deleted.
        :param parser_name: The name of the parser that is handling the resource.
        :return:
        """

        for synonym in resource.all_strings():
            self.synonym_lookup[synonym.lower()].discard(resource)

        self.parser_to_curations[parser_name].discard(resource)
        # it may have already been popped, or not exist
        try:
            self.resource_to_parsers.pop(resource)
        except KeyError:
            pass

    def sync_resources(
        self,
        original_resource: Optional[OntologyStringResource],
        new_resource: OntologyStringResource,
        parser_name: str,
    ) -> None:
        """Synchronizes resources within the internal state.

        If an original resource is provided, it is removed from the resource
        dictionaries and the new resource is added. If no original resource is provided,
        only the new resource is added. Note that no action is taken if
        original_resource == new_resource

        :param original_resource: The original resource to be replaced. If None, no
            resource is replaced.
        :param new_resource: The new resource to be added.
        :param parser_name: The name of the parser that is handling the resource.
        :return:
        """
        resources_are_equal = False
        if original_resource:
            # Only do something if the resource has actually changed
            resources_are_equal = original_resource == new_resource
            if not resources_are_equal:
                self.delete_resource(original_resource, parser_name)
                self.parser_cache_invalid.add(parser_name)

        if not original_resource or not resources_are_equal:
            for synonym in new_resource.all_strings():
                self.synonym_lookup[synonym.lower()].add(new_resource)

            self.parser_to_curations[parser_name].add(new_resource)
            self.resource_to_parsers[new_resource].add(parser_name)
            self.parser_cache_invalid.add(parser_name)

    def save(self) -> Iterable[str]:
        """Saves updated resources to the model pack.

        :return:
        """

        for parser_name, curation_set in self.parser_to_curations.items():
            path = self.parser_to_path[parser_name]
            yield f"Saving updated resources to {path}"
            dump_ontology_string_resources(curation_set, path, force=True)

        self.clear_invalid_caches()

    @staticmethod
    def _load_pipeline() -> Pipeline:
        load_config.clear()  # type: ignore[attr-defined]
        cfg = load_config()
        pipeline: Pipeline = instantiate(cfg.Pipeline)
        return pipeline

    def clear_invalid_caches(self) -> None:
        for parser_name in set(self.parser_cache_invalid):
            parser = self.parsers[parser_name]
            logging.info("clearing invalid cache for %s", parser_name)
            parser.clear_cache()
            self.parser_cache_invalid.discard(parser_name)

    def load_cache_state_aware_pipeline(self) -> Pipeline:
        self.clear_invalid_caches()
        return self._load_pipeline()
